import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_series_equal

from wind_up.math_funcs import circ_diff

test_circ_diff_data = [
    (0, 0, 0),
    (2, 1, 1),
    (359, 1, -2),
    (90, 270, -180),
    (90, 270.1, 179.9),
    ([1, 90, 90], [-1, 270, 270.1], [2, -180, 179.9]),
    (pd.Series([1, 90, 90]), pd.Series([-1, 270, 270.1]), pd.Series([2, -180, 179.9])),
    (pd.Series([1, 359.1, 2.1]), 1, pd.Series([0, -1.9, 1.1])),
    (1, pd.Series([1, 359.1, 2.1]), pd.Series([0, 1.9, -1.1])),
]


@pytest.mark.parametrize(("angle1", "angle2", "expected"), test_circ_diff_data)
def test_circ_diff(angle1: float | np.generic, angle2: float | np.generic, expected: float | np.generic) -> None:
    if isinstance(expected, pd.Series):
        assert_series_equal(circ_diff(angle1, angle2), (expected))
    else:
        assert circ_diff(angle1, angle2) == pytest.approx(expected)


def test_within_bin() -> None:
    # this test replicates logic in detrend.py where an older version of circ_diff gets the wrong result
    d = 242
    dir_bin_width = 10.0
    assert not np.abs(circ_diff(d - 5, d)) < dir_bin_width / 2



from pathlib import Path

import pandas as pd

from wind_up.caching import with_parquet_cache


class TestWithParquetCache:
    def test_creates_and_fetches_cache(self, tmp_path: Path) -> None:
        fp = tmp_path / "test.parquet"
        sample_df = pd.DataFrame({"a": [1, 2, 3]})

        @with_parquet_cache(fp=tmp_path / "test.parquet")
        def myfunc() -> pd.DataFrame:
            return sample_df

        assert not fp.exists()
        _df = myfunc()
        pd.testing.assert_frame_equal(_df, sample_df)
        assert fp.exists()

        df2 = myfunc()
        pd.testing.assert_frame_equal(df2, sample_df)

    def test_doesnt_run_the_func_if_file_exists(self, tmp_path: Path) -> None:
        fp = tmp_path / "test.parquet"
        sample_df = pd.DataFrame({"a": [1, 2, 3]})
        sample_df.to_parquet(fp)

        @with_parquet_cache(fp=tmp_path / "test.parquet")
        def myfunc() -> pd.DataFrame:
            return 1 / 0

        _df = myfunc()
        pd.testing.assert_frame_equal(_df, sample_df)



import matplotlib.pyplot as plt
import pandas as pd

from wind_up.models import PlotConfig


def plot_rolling_windspeed_diff_one_wtg(
    *,
    wtg_df: pd.DataFrame,
    wtg_name: str,
    ws_col: str,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure()
    plt.plot(wtg_df["rolling_windspeed_diff"])
    plot_title = f"{wtg_name} rolling {ws_col} diff to reanalysis"
    plt.title(plot_title)
    plt.xlabel("datetime")
    plt.ylabel("rolling_windspeed_diff [m/s]")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        (plot_cfg.plots_dir / wtg_name).mkdir(exist_ok=True, parents=True)
        plt.savefig(plot_cfg.plots_dir / wtg_name / f"{plot_title}.png")
    plt.close()



import pandas as pd

from wind_up.constants import RAW_DOWNTIME_S_COL, RAW_POWER_COL

YAW_OK_PW_FRACTION = 0.05


def add_ok_yaw_col(
    wf_or_wtg_df: pd.DataFrame, *, new_col_name: str, wd_col: str, rated_power: float, timebase_s: int
) -> pd.DataFrame:
    yaw_of_downtime_s = timebase_s * 1 / 4
    wf_or_wtg_df[new_col_name] = (
        wf_or_wtg_df[wd_col].notna()
        & (wf_or_wtg_df[RAW_POWER_COL] > rated_power * YAW_OK_PW_FRACTION)
        & (wf_or_wtg_df[RAW_DOWNTIME_S_COL].fillna(0) < yaw_of_downtime_s)
    )
    return wf_or_wtg_df



from typing import TypeVar

import pandas as pd

T = TypeVar("T", pd.Timestamp, pd.DatetimeIndex)


def ensure_utc(t: T) -> T:
    if t.tzinfo is None:
        return t.tz_localize("UTC")
    return t.tz_convert("UTC")



import logging
import math
from dataclasses import dataclass

import pandas as pd

from wind_up.constants import (
    REANALYSIS_WD_COL,
    REANALYSIS_WS_COL,
    TIMESTAMP_COL,
)
from wind_up.models import PlotConfig, WindUpConfig
from wind_up.plots.reanalysis_plots import plot_find_best_shift_and_corr, plot_wf_and_reanalysis_sample_timeseries

logger = logging.getLogger(__name__)


def reanalysis_upsample(reanalysis_raw_df: pd.DataFrame, *, timebase_s: int) -> pd.DataFrame:
    reanalysis_df = reanalysis_raw_df.resample(pd.Timedelta(seconds=timebase_s), label="left").last()
    upsample_factor = round(len(reanalysis_df) / len(reanalysis_raw_df))
    # extend the end of the index e.g. by 50 minutes for 10 minute timebase
    reanalysis_df = pd.concat(
        [
            reanalysis_df,
            pd.DataFrame(
                index=pd.date_range(
                    start=reanalysis_df.index[-1] + pd.Timedelta(seconds=timebase_s),
                    periods=upsample_factor - 1,
                    freq=pd.Timedelta(seconds=timebase_s),
                ),
            ),
        ],
    )
    reanalysis_df = reanalysis_df.ffill(limit=upsample_factor - 1)
    if "100_m_hws_mean_mps" in reanalysis_df.columns and "100_m_hwd_mean_deg-n_true" in reanalysis_df.columns:
        reanalysis_df = reanalysis_df.rename(
            columns={"100_m_hws_mean_mps": REANALYSIS_WS_COL, "100_m_hwd_mean_deg-n_true": REANALYSIS_WD_COL},
        )
    elif "50_m_hws_mean_mps" in reanalysis_df.columns and "50_m_hwd_mean_deg-n_true" in reanalysis_df.columns:
        reanalysis_df = reanalysis_df.rename(
            columns={"50_m_hws_mean_mps": REANALYSIS_WS_COL, "50_m_hwd_mean_deg-n_true": REANALYSIS_WD_COL},
        )
    else:
        msg = "reanalysis wind speed and direction columns not found"
        raise RuntimeError(msg)
    reanalysis_df.index.name = TIMESTAMP_COL
    return reanalysis_df


def find_best_shift_and_corr(
    *,
    wf_ws_df: pd.DataFrame,
    reanalysis_df: pd.DataFrame,
    wf_name: str,
    datastream_id: str,
    timebase_s: int,
    plot_cfg: PlotConfig | None,
    wf_ws_lower_limit: float = 0,
) -> tuple[float, int]:
    ws_filt_df = wf_ws_df["WindSpeedMean"][wf_ws_df["WindSpeedMean"] >= wf_ws_lower_limit].to_frame()
    best_s = -99
    best_corr = -99.0
    shifts = []
    corrs = []
    rows_per_hour = 3600 / timebase_s
    for s in range(round(-24 * rows_per_hour), round(24 * rows_per_hour), math.ceil(rows_per_hour / 6)):
        this_corr = float(ws_filt_df.corrwith(reanalysis_df[REANALYSIS_WS_COL].shift(s)).squeeze())
        shifts.append(s)
        corrs.append(this_corr)
        if this_corr > best_corr:
            best_corr = this_corr
            best_s = s
    if plot_cfg is not None:
        plot_find_best_shift_and_corr(
            wf_ws_df=ws_filt_df,
            reanalysis_df=reanalysis_df,
            shifts=shifts,
            corrs=corrs,
            wf_name=wf_name,
            datastream_id=datastream_id,
            best_corr=best_corr,
            best_s=best_s,
            timebase_s=timebase_s,
            plot_cfg=plot_cfg,
        )

    logger.info(f"{datastream_id} best correlation is {best_corr:.6f} with a shift of {best_s}")

    return best_corr, best_s


def calc_wf_mean_wind_speed_df(
    wf_df: pd.DataFrame,
    *,
    num_turbines: int,
    allowed_data_coverage_width: float,
) -> pd.DataFrame:
    wf_ws_df = wf_df.groupby(TIMESTAMP_COL).agg(
        WindSpeedMean=pd.NamedAgg(column="WindSpeedMean", aggfunc=lambda x: x.mean()),
        data_coverage=pd.NamedAgg(column="WindSpeedMean", aggfunc=lambda x: x.count() / num_turbines),
    )
    median_data_coverage = wf_ws_df["data_coverage"].median()
    wf_ws_df = wf_ws_df[(wf_ws_df["data_coverage"] - median_data_coverage).abs() < allowed_data_coverage_width / 2]
    return wf_ws_df.dropna()


def get_dsid_and_dates_from_filename(filename: str) -> tuple[str, pd.Timestamp, pd.Timestamp]:
    fname = filename.replace(".parquet", "")
    date_to = fname.split("_")[-1]
    date_from = fname.split("_")[-2]
    dsid = "_".join(fname.split("_")[:-2])
    return (
        dsid,
        pd.to_datetime(date_from, format="%Y%m%d").tz_localize("UTC"),
        pd.to_datetime(date_to, format="%Y%m%d").tz_localize("UTC"),
    )


@dataclass
class ReanalysisDataset:
    id: str
    data: pd.DataFrame


def add_reanalysis_data(
    wf_df: pd.DataFrame,
    *,
    reanalysis_datasets: list[ReanalysisDataset],
    cfg: WindUpConfig,
    plot_cfg: PlotConfig | None,
    require_full_coverage: bool = False,
) -> pd.DataFrame:
    data_coverage_width = 0.1
    wf_ws_df = calc_wf_mean_wind_speed_df(
        wf_df,
        num_turbines=len(cfg.asset.wtgs),
        allowed_data_coverage_width=data_coverage_width,
    )
    max_data_coverage_width = 0.3
    rows_per_hour = 3600 / cfg.timebase_s
    while len(wf_ws_df) < (60 * 24 * rows_per_hour) and data_coverage_width < max_data_coverage_width:
        data_coverage_width += 0.05
        wf_ws_df = calc_wf_mean_wind_speed_df(
            wf_df,
            num_turbines=len(cfg.asset.wtgs),
            allowed_data_coverage_width=data_coverage_width,
        )

    best_id = None
    best_df = None
    best_corr = -99.0
    best_s = None
    for reanalysis_dataset in reanalysis_datasets:
        _starts_later = reanalysis_dataset.data.index.min() > cfg.lt_first_dt_utc_start
        _ends_earlier = reanalysis_dataset.data.index.max() < cfg.analysis_last_dt_utc_start
        if (_starts_later or _ends_earlier) and require_full_coverage:
            continue  # skip to next dataset

        dsid = reanalysis_dataset.id
        this_reanalysis_df = reanalysis_upsample(reanalysis_dataset.data, timebase_s=cfg.timebase_s)

        this_corr, this_s = find_best_shift_and_corr(
            wf_ws_df=wf_ws_df,
            reanalysis_df=this_reanalysis_df,
            wf_name=cfg.asset.name,
            datastream_id=dsid,
            timebase_s=cfg.timebase_s,
            plot_cfg=plot_cfg,
        )
        if this_corr > best_corr:
            best_id = dsid
            best_df = this_reanalysis_df
            best_corr = this_corr
            best_s = this_s

    if best_df is None:
        msg = "no best_id found"
        raise RuntimeError(msg)

    logger.info(f"{best_id} has best correlation: {best_corr:.3f} with a shift of {best_s}")

    wf_and_reanalysis_df = wf_df.merge(
        best_df.shift(best_s)[[REANALYSIS_WS_COL, REANALYSIS_WD_COL]],
        how="left",
        left_index=True,
        right_index=True,
    )
    if plot_cfg is not None:
        plot_wf_and_reanalysis_sample_timeseries(wf_df=wf_and_reanalysis_df, plot_cfg=plot_cfg)
    return wf_and_reanalysis_df



import datetime as dt
import logging
import warnings
from typing import TypeAlias

import numpy as np
import pandas as pd

from wind_up.constants import RAW_DOWNTIME_S_COL, RAW_POWER_COL, RAW_WINDSPEED_COL, RAW_YAWDIR_COL, DataColumns
from wind_up.math_funcs import circ_diff
from wind_up.models import PlotConfig, Turbine, WindUpConfig
from wind_up.plots.scada_funcs_plots import (
    plot_data_coverage_heatmap,
    plot_filter_rpm_and_pt_curve_one_ttype_or_wtg,
    plot_ops_curves_per_ttype,
    print_and_plot_capacity_factor,
    print_filter_stats,
)
from wind_up.smart_data import add_smart_lat_long_to_cfg, load_smart_scada_and_md_from_file

logger = logging.getLogger(__name__)
ExclusionPeriodsType: TypeAlias = list[tuple[str, dt.datetime, dt.datetime]]


def filter_stuck_data(df: pd.DataFrame) -> pd.DataFrame:
    diffdf = df.groupby("TurbineName", observed=False).ffill().fillna(0).diff()
    stuck_data = (diffdf == 0).all(axis=1)
    very_low_wind_threshold = 1.5
    very_low_wind = df["WindSpeedMean"] < very_low_wind_threshold
    stuck_filter = stuck_data & (~very_low_wind)
    df.loc[stuck_filter, :] = pd.NA
    print_filter_stats(filter_name="filter_stuck_data", na_rows=stuck_filter.sum(), total_rows=len(df))
    return df


def filter_bad_pw_ws(df: pd.DataFrame, max_rated_power: float) -> pd.DataFrame:
    na_rows_before = df["ActivePowerMean"].isna().sum()
    df.loc[df[["ActivePowerMean", "WindSpeedMean"]].isna().any(axis=1), :] = pd.NA
    df.loc[(df["ActivePowerMean"] < -0.5 * max_rated_power), :] = pd.NA
    df.loc[(df["ActivePowerMean"] > 2 * max_rated_power), :] = pd.NA
    df.loc[(df["WindSpeedMean"] < 0), :] = pd.NA
    ws_in_range_ul = 98
    df.loc[(df["WindSpeedMean"] > ws_in_range_ul), :] = pd.NA
    na_rows_after = df["ActivePowerMean"].isna().sum()
    na_rows = na_rows_after - na_rows_before
    print_filter_stats(filter_name="filter_bad_pw_ws", na_rows=na_rows, total_rows=len(df))
    return df


def wrap_yaw_and_pitch(df: pd.DataFrame) -> pd.DataFrame:
    yaw_lt0 = (df["YawAngleMean"] < 0).sum()
    yaw_ge360 = (df["YawAngleMean"] >= 360).sum()  # noqa PLR2004
    if yaw_lt0 > 0 or yaw_ge360 > 0:
        logger.info(f"rows with YawAngleMean lt 0: {yaw_lt0}")
        logger.info(f"rows with YawAngleMean ge 360: {yaw_ge360}")
        df["YawAngleMean"] = df["YawAngleMean"] % 360
        logger.info("wrapped YawAngleMean to 0-360")
        if "YawAngleMin" in df.columns:
            df["YawAngleMin"] = df["YawAngleMin"].clip(lower=0, upper=360) % 360
            logger.info("clipped YawAngleMin to 0-360")
        if "YawAngleMax" in df.columns:
            df["YawAngleMax"] = df["YawAngleMax"].clip(lower=0, upper=360) % 360
            logger.info("clipped YawAngleMax to 0-360")

    # ensure pitch data is -180 to 180
    pt_lt_180 = (df["PitchAngleMean"] < -180).sum()  # noqa PLR2004
    pt_ge_180 = (df["PitchAngleMean"] >= 180).sum()  # noqa PLR2004
    if pt_lt_180 > 0 or pt_ge_180 > 0:
        logger.info(f"rows with PitchAngleMean lt -180: {pt_lt_180}")
        logger.info(f"rows with PitchAngleMean ge 180: {pt_ge_180}")
        df["PitchAngleMean"] = ((df["PitchAngleMean"] + 180) % 360) - 180
        logger.info("wrapped PitchAngleMean to -180-180")
    return df


def filter_wrong_yaw(df: pd.DataFrame) -> pd.DataFrame:
    if "YawAngleMin" in df.columns and "YawAngleMax" in df.columns:
        min_wrong = (df["YawAngleMin"] > df["YawAngleMean"]) | (df["YawAngleMin"] > df["YawAngleMax"])
        max_wrong = (df["YawAngleMax"] < df["YawAngleMean"]) | (df["YawAngleMax"] < df["YawAngleMin"])
        min_max_filter = min_wrong | max_wrong
        df.loc[min_max_filter, ["YawAngleMin", "YawAngleMax"]] = pd.NA
        print_filter_stats(
            filter_name="filter_wrong_yaw",
            na_rows=min_max_filter.sum(),
            total_rows=len(df),
            just_yaw=True,
            just_min_max=True,
            reason="bad YawAngleMin/Max values",
        )

        # sometimes SCADA systems use arithmetic mean instead of circ mean
        wrap_rows = (df["YawAngleMin"] <= 1) & (df["YawAngleMax"] >= 359)  # noqa PLR2004
        mean_far_from_north = np.abs(circ_diff(df["YawAngleMean"].to_numpy(), 0)) >= 15  # noqa PLR2004
        bad_yaw_mean_filter = wrap_rows & mean_far_from_north
        yaw_cols = [col for col in df.columns if "Yaw" in col]
        df.loc[bad_yaw_mean_filter, yaw_cols] = pd.NA
        print_filter_stats(
            filter_name="filter_wrong_yaw",
            na_rows=bad_yaw_mean_filter.sum(),
            total_rows=len(df),
            just_yaw=True,
            reason="YawAngleMean appears to be wrong",
        )

    return df


def add_pw_clipped(df: pd.DataFrame, wtgs: list[Turbine]) -> pd.DataFrame:
    df["pw_clipped"] = df["ActivePowerMean"].clip(lower=0)
    for i, rp in zip(
        [x.name for x in wtgs],
        [x.turbine_type.rated_power_kw for x in wtgs],
        strict=True,
    ):
        df.loc[i, "pw_clipped"] = df.loc[i, "pw_clipped"].clip(upper=rp).to_numpy()
    return df


def filter_exclusions(
    input_df: pd.DataFrame,
    exclusion_periods_utc: ExclusionPeriodsType,
) -> pd.DataFrame:
    pw_na_before = input_df["ActivePowerMean"].isna().sum()
    for exclusion in exclusion_periods_utc:
        excl_turbine = exclusion[0]
        exclusion_period_start_utc = exclusion[1]
        exclusion_period_end_utc = exclusion[2]
        if excl_turbine == "ALL":
            input_df.loc[
                pd.IndexSlice[
                    :,
                    (exclusion_period_start_utc - pd.Timedelta(minutes=9, seconds=59)) : (
                        exclusion_period_end_utc - pd.Timedelta(seconds=1)
                    ),
                ],
                :,
            ] = pd.NA
        else:
            input_df.loc[
                pd.IndexSlice[
                    excl_turbine,
                    (exclusion_period_start_utc - pd.Timedelta(minutes=9, seconds=59)) : (
                        exclusion_period_end_utc - pd.Timedelta(seconds=1)
                    ),
                ],
                :,
            ] = pd.NA
    pw_na_after = input_df["ActivePowerMean"].isna().sum()
    print_filter_stats(filter_name="filter_exclusions", na_rows=pw_na_after - pw_na_before, total_rows=len(input_df))
    # also filter any row where "exclude_row" is True
    if "exclude_row" in input_df.columns:
        pw_na_before = input_df["ActivePowerMean"].isna().sum()
        input_df.loc[input_df["exclude_row"].fillna(value=0) == 1, :] = pd.NA
        pw_na_after = input_df["ActivePowerMean"].isna().sum()
        print_filter_stats(
            filter_name="filter_exclusions exclude_row", na_rows=pw_na_after - pw_na_before, total_rows=len(input_df)
        )
    return input_df


def filter_yaw_exclusions(
    input_df: pd.DataFrame,
    yaw_data_exclusions_utc: ExclusionPeriodsType,
) -> pd.DataFrame:
    yaw_cols = [col for col in input_df.columns if "Yaw" in col]
    yaw_na_before = input_df["YawAngleMean"].isna().sum()
    for exclusion in yaw_data_exclusions_utc:
        excl_turbine = exclusion[0]
        exclusion_period_start_utc = exclusion[1]
        exclusion_period_end_utc = exclusion[2]
        if excl_turbine == "ALL":
            input_df.loc[
                pd.IndexSlice[
                    :,
                    (exclusion_period_start_utc - pd.Timedelta(minutes=9, seconds=59)) : (
                        exclusion_period_end_utc - pd.Timedelta(seconds=1)
                    ),
                ],
                yaw_cols,
            ] = pd.NA
        else:
            input_df.loc[
                pd.IndexSlice[
                    excl_turbine,
                    (exclusion_period_start_utc - pd.Timedelta(minutes=9, seconds=59)) : (
                        exclusion_period_end_utc - pd.Timedelta(seconds=1)
                    ),
                ],
                yaw_cols,
            ] = pd.NA
    yaw_na_after = input_df["YawAngleMean"].isna().sum()
    print_filter_stats(
        filter_name="filter_yaw_exclusions",
        na_rows=yaw_na_after - yaw_na_before,
        total_rows=len(input_df),
        just_yaw=True,
    )
    return input_df


def filter_cfg_exclusions(
    input_df: pd.DataFrame, exclusion_periods_utc: ExclusionPeriodsType, yaw_data_exclusions_utc: ExclusionPeriodsType
) -> pd.DataFrame:
    input_df = filter_exclusions(input_df=input_df, exclusion_periods_utc=exclusion_periods_utc)
    return filter_yaw_exclusions(input_df=input_df, yaw_data_exclusions_utc=yaw_data_exclusions_utc)


def filter_downtime(df: pd.DataFrame) -> pd.DataFrame:
    pw_na_before = df["ActivePowerMean"].isna().sum()
    fully_available = df["ShutdownDuration"] == 0
    df.loc[~fully_available, :] = pd.NA
    pw_na_after = df["ActivePowerMean"].isna().sum()
    print_filter_stats(filter_name="filter_downtime", na_rows=pw_na_after - pw_na_before, total_rows=len(df))
    return df


def filter_rpm_and_pt_oor_one_ttype(
    df: pd.DataFrame,
    rpm_lower: float,
    rpm_upper: float,
    pt_lower: float,
    pt_upper: float,
) -> tuple[pd.DataFrame, int]:
    na_rows_before = df["ActivePowerMean"].isna().sum()
    df.loc[(df["GenRpmMean"] < rpm_lower), :] = pd.NA
    df.loc[(df["GenRpmMean"] > rpm_upper), :] = pd.NA
    df.loc[(df["PitchAngleMean"] < pt_lower), :] = pd.NA
    df.loc[(df["PitchAngleMean"] > pt_upper), :] = pd.NA
    na_rows_after = df["ActivePowerMean"].isna().sum()
    na_rows = na_rows_after - na_rows_before
    return df, na_rows


def filter_rpm_or_pt_curve(
    df: pd.DataFrame,
    *,
    x_col: str,
    y_col: str,
    y_centile: float,
    x_bin_edges: list[float],
    reject_high: bool,
    y_margin: float,
    filter_last_bin: bool = True,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    # suppress RuntimeWarning due to empty slices, for example the high wind speed bins are often empty.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        filter_curve = df.groupby(by=pd.cut(df[x_col], bins=x_bin_edges, retbins=False), observed=False).agg(
            y_limit=pd.NamedAgg(column=y_col, aggfunc=lambda x: np.nanpercentile(x, y_centile)),
            min_y=pd.NamedAgg(column=y_col, aggfunc=lambda x: x.min()),
            max_y=pd.NamedAgg(column=y_col, aggfunc=lambda x: x.max()),
        )

    margin_sign = 1 if reject_high else -1
    filter_curve["y_limit"] = filter_curve["y_limit"] + (margin_sign * y_margin)

    if not filter_last_bin:
        if reject_high:
            filter_curve.loc[filter_curve.index[-1], "y_limit"] = filter_curve.iloc[-1]["max_y"]
        else:
            filter_curve.loc[filter_curve.index[-1], "y_limit"] = filter_curve.iloc[-1]["min_y"]

    filt_threshold = np.interp(df[x_col].values, [x.mid for x in filter_curve.index], filter_curve["y_limit"])
    if reject_high:
        df.loc[df[y_col] > filt_threshold, :] = pd.NA
    else:
        df.loc[df[y_col] < filt_threshold, :] = pd.NA

    return df, filter_curve


def filter_rpm_and_pt_curve_one_ttype(
    df: pd.DataFrame,
    *,
    ttype: str,
    pitch_to_stall: bool,
    rated_power_kw: float,
    rpm_v_pw_margin_factor: float,
    plot_cfg: PlotConfig | None,
) -> pd.DataFrame:
    na_rows_before = df["ActivePowerMean"].isna().sum()
    df_pre_filter = df.copy()

    if pitch_to_stall:
        pt_centile_pw = 10
        pt_centile_ws = 1
    else:
        pt_centile_pw = 90
        pt_centile_ws = 99
    rpm_centile_pw = 10
    rpm_centile_ws = 1
    pt_margin_pw = 2.5
    pt_margin_ws = 2.5
    rpm_margin_pw = df["GenRpmMean"].max() * rpm_v_pw_margin_factor
    rpm_margin_ws = df["GenRpmMean"].max() * 0.05

    pw_bin_width = rated_power_kw / 50
    pw_bin_edges = list(np.arange(0, rated_power_kw + pw_bin_width, pw_bin_width))

    ws_bin_width = 2
    ws_bin_edges = list(np.arange(0, df["WindSpeedMean"].max() + ws_bin_width, ws_bin_width))

    for _x in range(2):
        # the below filter steps are performed twice in case there is interaction between the filters
        df, pt_v_pw_curve = filter_rpm_or_pt_curve(
            df=df,
            x_col="pw_clipped",
            y_col="PitchAngleMean",
            y_centile=pt_centile_pw,
            x_bin_edges=pw_bin_edges,
            reject_high=not pitch_to_stall,
            y_margin=pt_margin_pw,
            filter_last_bin=False,
        )

        df, pt_v_ws_curve = filter_rpm_or_pt_curve(
            df=df,
            x_col="WindSpeedMean",
            y_col="PitchAngleMean",
            y_centile=pt_centile_ws,
            x_bin_edges=ws_bin_edges,
            reject_high=not pitch_to_stall,
            y_margin=pt_margin_ws,
        )

        df, rpm_v_pw_curve = filter_rpm_or_pt_curve(
            df=df,
            x_col="pw_clipped",
            y_col="GenRpmMean",
            y_centile=rpm_centile_pw,
            x_bin_edges=pw_bin_edges,
            reject_high=False,
            y_margin=rpm_margin_pw,
        )

        df, rpm_v_ws_curve = filter_rpm_or_pt_curve(
            df=df,
            x_col="WindSpeedMean",
            y_col="GenRpmMean",
            y_centile=rpm_centile_ws,
            x_bin_edges=ws_bin_edges,
            reject_high=False,
            y_margin=rpm_margin_ws,
        )

    if plot_cfg is not None:
        plot_filter_rpm_and_pt_curve_one_ttype_or_wtg(
            df=df_pre_filter,
            ttype_or_wtg=ttype,
            pt_v_pw_curve=pt_v_pw_curve,
            pt_v_ws_curve=pt_v_ws_curve,
            rpm_v_pw_curve=rpm_v_pw_curve,
            rpm_v_ws_curve=rpm_v_ws_curve,
            plot_cfg=plot_cfg,
        )
        if not plot_cfg.skip_per_turbine_plots:
            for wtg in df_pre_filter.index.unique(level="TurbineName"):
                plot_filter_rpm_and_pt_curve_one_ttype_or_wtg(
                    df=df_pre_filter.loc[wtg],
                    ttype_or_wtg=wtg,
                    pt_v_pw_curve=pt_v_pw_curve,
                    pt_v_ws_curve=pt_v_ws_curve,
                    rpm_v_pw_curve=rpm_v_pw_curve,
                    rpm_v_ws_curve=rpm_v_ws_curve,
                    plot_cfg=plot_cfg,
                )

    na_rows_after = df["ActivePowerMean"].isna().sum()
    na_rows = na_rows_after - na_rows_before
    return df, na_rows


def filter_missing_rpm_or_pt(df: pd.DataFrame) -> pd.DataFrame:
    pw_na_before = df["ActivePowerMean"].isna().sum()
    df.loc[df[["PitchAngleMean", "GenRpmMean"]].isna().any(axis=1), :] = pd.NA
    pw_na_after = df["ActivePowerMean"].isna().sum()
    print_filter_stats(
        filter_name="filter_missing_rpm_and_pt",
        na_rows=pw_na_after - pw_na_before,
        total_rows=len(df),
        reason="rpm or pitch are NA",
    )
    return df


def filter_rpm_and_pt(input_df: pd.DataFrame, cfg: WindUpConfig, plot_cfg: PlotConfig | None) -> pd.DataFrame:
    df_idx_before = input_df.index.copy()

    input_df = filter_missing_rpm_or_pt(df=input_df)

    # filter out of range pitch and rpm by turbine type
    df_prefilt = input_df.copy()
    filt_df = pd.DataFrame()
    na_rows = 0
    for ttype in cfg.list_unique_turbine_types():
        wtgs = cfg.list_turbine_ids_of_type(ttype)
        df_ttype = df_prefilt.loc[wtgs]
        rpm_lower, rpm_upper = cfg.get_normal_operation_genrpm_range(ttype=ttype)
        pt_lower, pt_upper = cfg.get_normal_operation_pitch_range(ttype=ttype)
        df_, na_rows_ = filter_rpm_and_pt_oor_one_ttype(
            df=df_ttype,
            rpm_lower=rpm_lower,
            rpm_upper=rpm_upper,
            pt_lower=pt_lower,
            pt_upper=pt_upper,
        )
        filt_df = pd.concat([filt_df, df_])
        na_rows += na_rows_
    filt_df = filt_df.sort_index()
    print_filter_stats(
        filter_name="filter_rpm_and_pt",
        na_rows=na_rows,
        total_rows=len(filt_df),
        reason="rpm or pitch are out of range",
    )

    # filter based on ops curves by turbine type
    df_prefilt = filt_df.copy()
    filt_df = pd.DataFrame()
    na_rows = 0
    for ttype in cfg.list_unique_turbine_types():
        wtgs = cfg.list_turbine_ids_of_type(ttype)
        df_ttype = df_prefilt.loc[wtgs]
        df_, na_rows_ = filter_rpm_and_pt_curve_one_ttype(
            df=df_ttype,
            ttype=ttype.turbine_type,
            pitch_to_stall=ttype.pitch_to_stall,
            rated_power_kw=ttype.rated_power_kw,
            rpm_v_pw_margin_factor=ttype.rpm_v_pw_margin_factor,
            plot_cfg=plot_cfg,
        )
        filt_df = pd.concat([filt_df, df_])
        na_rows += na_rows_
    filt_df = filt_df.sort_index()
    print_filter_stats(
        filter_name="filter_rpm_and_pt",
        na_rows=na_rows,
        total_rows=len(filt_df),
        reason="rpm or pitch curve filtering",
    )

    # check df.index is the same as df_idx_before
    if not filt_df.index.equals(df_idx_before):
        msg = "df.index has changed during filtering"
        raise ValueError(msg)
    return filt_df


def scada_multi_index(df: pd.DataFrame) -> pd.DataFrame:
    return df.set_index("TurbineName", append=True).swaplevel(axis=0).sort_index()


def filter_scada_df(raw_df: pd.DataFrame, cfg: WindUpConfig, plot_cfg: PlotConfig) -> pd.DataFrame:
    ref_col = DataColumns.active_power_mean
    logger.info(
        f"{raw_df[ref_col].isna().sum()} rows "
        f"[{100 * raw_df[ref_col].isna().mean():.1f}%] "
        "of power data is missing before filtering",
    )
    filt_df = raw_df.copy()
    filt_df = filter_stuck_data(df=filt_df)
    filt_df = filter_bad_pw_ws(df=filt_df, max_rated_power=cfg.get_max_rated_power())
    filt_df = wrap_yaw_and_pitch(df=filt_df)
    filt_df = filter_wrong_yaw(df=filt_df)
    filt_df = filter_cfg_exclusions(
        input_df=filt_df,
        exclusion_periods_utc=cfg.exclusion_periods_utc,
        yaw_data_exclusions_utc=cfg.yaw_data_exclusions_utc,
    )
    filt_df = filter_downtime(df=filt_df)
    plot_ops_curves_per_ttype(cfg=cfg, df=filt_df, title_end="downtime filter", plot_cfg=plot_cfg)
    filt_df = add_pw_clipped(df=filt_df, wtgs=cfg.asset.wtgs)
    filt_df = filter_rpm_and_pt(input_df=filt_df, cfg=cfg, plot_cfg=plot_cfg)
    plot_data_coverage_heatmap(
        df=filt_df,
        plot_title=f"{cfg.asset.name} data coverage after filtering",
        plot_cfg=plot_cfg,
    )
    logger.info(
        f"{filt_df[ref_col].isna().sum()} rows [{100 * filt_df[ref_col].isna().mean():.1f}%] "
        "of power data is missing after filtering",
    )

    filt_df[RAW_POWER_COL] = raw_df[DataColumns.active_power_mean]
    filt_df[RAW_WINDSPEED_COL] = raw_df[DataColumns.wind_speed_mean]
    filt_df[RAW_DOWNTIME_S_COL] = raw_df[DataColumns.shutdown_duration]
    filt_df[RAW_YAWDIR_COL] = raw_df[DataColumns.yaw_angle_mean]

    plot_ops_curves_per_ttype(cfg=cfg, df=filt_df, title_end="after filtering", plot_cfg=plot_cfg)

    return filt_df


def get_raw_scada_and_cfg_from_file(
    cfg: WindUpConfig,
    scada_df: pd.DataFrame,
    metadata_df: pd.DataFrame,
    first_datetime_utc_start: pd.Timestamp,
    last_datetime_utc_start: pd.Timestamp,
    plot_cfg: PlotConfig | None,
) -> tuple[pd.DataFrame, WindUpConfig]:
    scada_raw, md = load_smart_scada_and_md_from_file(
        asset_name=cfg.asset.name,
        scada_df=scada_df,
        metadata_df=metadata_df,
        first_datetime_utc_start=first_datetime_utc_start,
        last_datetime_utc_start=last_datetime_utc_start,
        timebase_s=cfg.timebase_s,
    )
    cfg = add_smart_lat_long_to_cfg(md=md, cfg=cfg)
    if plot_cfg is not None:
        print_and_plot_capacity_factor(scada_df=scada_raw, cfg=cfg, plots_cfg=plot_cfg)
    scada_mi_df = scada_multi_index(scada_raw)
    del scada_raw
    if plot_cfg is not None:
        plot_ops_curves_per_ttype(cfg=cfg, df=scada_mi_df, title_end="before filtering", plot_cfg=plot_cfg)
        plot_data_coverage_heatmap(
            df=scada_mi_df,
            plot_title=f"{cfg.asset.name} data coverage before filtering",
            plot_cfg=plot_cfg,
        )
    return scada_mi_df, cfg



import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from wind_up.math_funcs import circ_diff
from wind_up.models import PlotConfig


def plot_yaw_direction_pre_post_per_signal(
    signal_name: str,
    test_wd_col: str,
    *,
    plot_pre_df: pd.DataFrame,
    plot_post_df: pd.DataFrame,
    test_name: str,
    ref_name: str,
    ref_ws_col: str,
    ref_wd_col: str,
    plot_cfg: PlotConfig,
    toggle_name: str | None = None,
) -> None:
    pre_label = f"{toggle_name} OFF" if toggle_name else "pre upgrade"
    post_label = f"{toggle_name} ON" if toggle_name else "post upgrade"

    plt.figure(figsize=(12, 8))
    plt.subplot(2, 1, 1)
    plt.scatter(
        plot_pre_df[ref_wd_col],
        plot_pre_df[test_wd_col],
        s=2,
        alpha=0.5,
        label=pre_label,
    )
    plt.scatter(
        plot_post_df[ref_wd_col],
        plot_post_df[test_wd_col],
        s=2,
        alpha=0.5,
        label=post_label,
    )
    plt.xlabel(f"{ref_wd_col} [deg]")
    plt.ylabel(f"{test_wd_col} [deg]")
    plt.grid()
    plt.legend(loc="best", markerscale=2)

    plt.subplot(2, 1, 2)
    plt.scatter(
        plot_pre_df[ref_wd_col],
        plot_pre_df[signal_name],
        s=2,
        alpha=0.5,
        label=pre_label,
    )
    plt.scatter(
        plot_post_df[ref_wd_col],
        plot_post_df[signal_name],
        s=2,
        alpha=0.5,
        label=post_label,
    )
    plt.xlabel(f"{ref_wd_col} [deg]")
    plt.ylabel(f'{signal_name.replace("_", " ")} [deg]')
    plt.grid()

    plot_title = (
        f"{test_name} ref {ref_name} yaw direction scatter"
        if signal_name == "yaw_offset"
        else f'{test_name} ref {ref_name} {signal_name.replace("_", " ")} scatter'
    )
    plt.suptitle(plot_title)
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        (plot_cfg.plots_dir / test_name / "yaw_direction").mkdir(exist_ok=True)
        plt.savefig(plot_cfg.plots_dir / test_name / "yaw_direction" / f"{plot_title}.png")
    plt.close()

    ws_bin_width = 2
    wd_bin_width = 5

    ws_bin_edges = np.arange(
        0, max(plot_pre_df[ref_ws_col].max(), plot_post_df[ref_ws_col].max()) + ws_bin_width, ws_bin_width
    )
    plot_pre_df["ws_bins"] = pd.cut(plot_pre_df[ref_ws_col], bins=ws_bin_edges, retbins=False)
    plot_pre_df["ws_bin_centre"] = [x.mid for x in plot_pre_df["ws_bins"]]
    plot_post_df["ws_bins"] = pd.cut(plot_post_df[ref_ws_col], bins=ws_bin_edges, retbins=False)
    plot_post_df["ws_bin_centre"] = [x.mid for x in plot_post_df["ws_bins"]]

    def _get_mid(x: float | pd.Interval) -> float:
        if isinstance(x, pd.Interval):
            return x.mid
        return x

    ws_bin_edges = np.arange(
        0, max(plot_pre_df[ref_ws_col].max(), plot_post_df[ref_ws_col].max()) + ws_bin_width, ws_bin_width
    )
    plot_pre_df["ws_bins"] = pd.cut(plot_pre_df[ref_ws_col], bins=ws_bin_edges, retbins=False)
    plot_pre_df["ws_bin_centre"] = plot_pre_df["ws_bins"].apply(_get_mid)
    plot_post_df["ws_bins"] = pd.cut(plot_post_df[ref_ws_col], bins=ws_bin_edges, retbins=False)
    plot_post_df["ws_bin_centre"] = plot_post_df["ws_bins"].apply(_get_mid)

    wd_bin_edges = np.arange(0, plot_pre_df[ref_wd_col].max() + wd_bin_width, wd_bin_width)
    plot_pre_df["wd_bins"] = pd.cut(plot_pre_df[ref_wd_col], bins=wd_bin_edges, retbins=False)
    plot_pre_df["wd_bin_centre"] = plot_pre_df["wd_bins"].apply(_get_mid)
    plot_post_df["wd_bins"] = pd.cut(plot_post_df[ref_wd_col], bins=wd_bin_edges, retbins=False)
    plot_post_df["wd_bin_centre"] = plot_post_df["wd_bins"].apply(_get_mid)

    for i, label in enumerate([pre_label, post_label]):
        plot_df = plot_pre_df if i == 0 else plot_post_df
        plt.figure(figsize=(8, 8))
        plt.subplot(2, 1, 1)
        sns.heatmap(
            plot_df.pivot_table(
                index="ws_bin_centre",
                columns="wd_bin_centre",
                values=signal_name,
                aggfunc=lambda x: x.count() / 6,
                observed=False,
            ).iloc[::-1],
            annot=True,
            cmap="gray_r",
            fmt=".1f",
            linewidths=0.5,
            cbar_kws={"label": "hours of data"},
        )
        plt.xlabel("wind direction bin centre [deg]")
        plt.ylabel("wind speed bin centre [m/s]")

        plt.subplot(2, 1, 2)
        sns.heatmap(
            plot_df.pivot_table(
                index="ws_bin_centre", columns="wd_bin_centre", values=signal_name, observed=False
            ).iloc[::-1],
            annot=True,
            cmap="YlGnBu",
            fmt=".1f",
            linewidths=0.5,
            vmin=0,
            vmax=20,
            cbar_kws={"label": f"{signal_name.replace('_',' ')} [deg]"},
        )

        plt.xlabel("wind direction bin centre [deg]")
        plt.ylabel("wind speed bin centre [m/s]")

        signal_descr = (
            f"{ref_name} minus {test_name} yaw direction"
            if signal_name == "yaw_offset"
            else f'{test_name} ref {ref_name} {signal_name.replace("_", " ")}'
        )
        plot_title = f"{signal_descr} vs ws and wd {label}"
        plt.suptitle(plot_title)
        plt.tight_layout()
        if plot_cfg.show_plots:
            plt.show()
        if plot_cfg.save_plots:
            (plot_cfg.plots_dir / test_name / "yaw_direction").mkdir(exist_ok=True)
            plt.savefig(plot_cfg.plots_dir / test_name / "yaw_direction" / f"{plot_title}.png")
        plt.close()


def plot_yaw_direction_pre_post(
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    *,
    test_name: str,
    ref_name: str,
    ref_ws_col: str,
    ref_wd_col: str,
    plot_cfg: PlotConfig,
    toggle_name: str | None = None,
) -> None:
    test_wd_col = "test_YawAngleMean"
    plot_pre_df = pre_df.dropna(subset=[test_wd_col, ref_ws_col, ref_wd_col]).copy()
    plot_post_df = post_df.dropna(subset=[test_wd_col, ref_ws_col, ref_wd_col]).copy()

    plot_pre_df["yaw_offset"] = circ_diff(plot_pre_df[ref_wd_col], plot_pre_df[test_wd_col])
    plot_post_df["yaw_offset"] = circ_diff(plot_post_df[ref_wd_col], plot_post_df[test_wd_col])

    plot_yaw_direction_pre_post_per_signal(
        signal_name="yaw_offset",
        test_wd_col=test_wd_col,
        plot_pre_df=plot_pre_df,
        plot_post_df=plot_post_df,
        test_name=test_name,
        ref_name=ref_name,
        ref_ws_col=ref_ws_col,
        ref_wd_col=ref_wd_col,
        plot_cfg=plot_cfg,
        toggle_name=toggle_name,
    )
    if "test_yaw_offset_command" in plot_pre_df.columns:
        plot_pre_df = plot_pre_df.rename(columns={"test_yaw_offset_command": "yaw_offset_command"})
        plot_post_df = plot_post_df.rename(columns={"test_yaw_offset_command": "yaw_offset_command"})
        plot_yaw_direction_pre_post_per_signal(
            signal_name="yaw_offset_command",
            test_wd_col=test_wd_col,
            plot_pre_df=plot_pre_df,
            plot_post_df=plot_post_df,
            test_name=test_name,
            ref_name=ref_name,
            ref_ws_col=ref_ws_col,
            ref_wd_col=ref_wd_col,
            plot_cfg=plot_cfg,
            toggle_name=toggle_name,
        )



import numpy as np
import numpy.typing as npt


def circ_diff(angle1: float | npt.NDArray | list, angle2: float | npt.NDArray | list) -> float | npt.NDArray:
    """Calculate the circular difference between two angles.

    Args:
        angle1: First angle in degrees.
        angle2: Second angle in degrees.

    Returns:
        Circular difference between the two angles in degrees
    """

    # Convert list to numpy array
    if isinstance(angle1, list):
        angle1 = np.array(angle1)
    if isinstance(angle2, list):
        angle2 = np.array(angle2)

    return np.mod(angle1 - angle2 + 180, 360) - 180



import numpy as np
import pandas as pd

from wind_up.models import PlotConfig, WindUpConfig
from wind_up.plots.windspeed_drift_plots import plot_rolling_windspeed_diff_one_wtg
from wind_up.result_manager import result_manager


def add_rolling_windspeed_diff(
    wtg_df: pd.DataFrame, *, ws_col: str, reanalysis_ws_col: str, timebase_s: int
) -> pd.DataFrame:
    wtg_df = wtg_df.copy()

    # check for ws drift issue
    wtg_df["ws_diff_to_renalysis"] = wtg_df[ws_col] - wtg_df[reanalysis_ws_col]
    ws_ll = 6
    ws_ul = 15
    wtg_df.loc[wtg_df[ws_col] < ws_ll, "ws_diff_to_renalysis"] = np.nan
    wtg_df.loc[wtg_df[ws_col] > ws_ul, "ws_diff_to_renalysis"] = np.nan
    rolling_days = 90
    rows_per_day = 24 * 3600 / timebase_s
    wtg_df["rolling_windspeed_diff"] = (
        wtg_df["ws_diff_to_renalysis"]
        .rolling(window=round(rolling_days * rows_per_day), min_periods=round(rolling_days * rows_per_day // 3))
        .median()
    )
    min_roll_days = 14
    while rolling_days >= (min_roll_days * 2) and len(wtg_df["rolling_windspeed_diff"].dropna()) == 0:
        rolling_days = rolling_days // 2
        wtg_df["rolling_windspeed_diff"] = (
            wtg_df["ws_diff_to_renalysis"]
            .rolling(window=round(rolling_days * rows_per_day), min_periods=round(rolling_days * rows_per_day // 3))
            .median()
        )
    if len(wtg_df["rolling_windspeed_diff"].dropna()) == 0:
        result_manager.warning("could not calculate rolling windspeed diff")
    return wtg_df


def calc_max_abs_relative_rolling_windspeed_diff(wtg_df: pd.DataFrame) -> float:
    return (wtg_df["rolling_windspeed_diff"] - wtg_df["rolling_windspeed_diff"].median()).abs().max()


def check_windspeed_drift(
    *,
    wtg_df: pd.DataFrame,
    wtg_name: str,
    ws_col: str,
    reanalysis_ws_col: str,
    cfg: WindUpConfig,
    plot_cfg: PlotConfig | None,
) -> tuple[float, float]:
    wtg_df = wtg_df.copy()
    wtg_df = add_rolling_windspeed_diff(
        wtg_df, ws_col=ws_col, reanalysis_ws_col=reanalysis_ws_col, timebase_s=cfg.timebase_s
    )
    if plot_cfg is not None:
        plot_rolling_windspeed_diff_one_wtg(wtg_df=wtg_df, wtg_name=wtg_name, ws_col=ws_col, plot_cfg=plot_cfg)

    max_abs_rel_diff = calc_max_abs_relative_rolling_windspeed_diff(wtg_df)
    max_abs_rel_diff_pp_period = calc_max_abs_relative_rolling_windspeed_diff(
        wtg_df.loc[cfg.analysis_first_dt_utc_start : cfg.analysis_last_dt_utc_start],  # type: ignore[misc]
    )

    ws_diff_ul = 0.5
    if max_abs_rel_diff > ws_diff_ul:
        result_manager.warning(f"possible wind speed drift of {max_abs_rel_diff:.1f} m/s for {wtg_name}")
    if max_abs_rel_diff_pp_period > ws_diff_ul:
        result_manager.warning(
            f"possible wind speed drift of {max_abs_rel_diff_pp_period:.1f} m/s for {wtg_name} "
            f"DURING POWER PERFORMANCE PERIOD",
        )

    return max_abs_rel_diff, max_abs_rel_diff_pp_period



import logging

import numpy as np
import pandas as pd
from scipy.stats import circmean

from wind_up.constants import (
    RAW_YAWDIR_COL,
    REANALYSIS_WD_COL,
    TIMESTAMP_COL,
    WINDFARM_YAWDIR_COL,
)
from wind_up.math_funcs import circ_diff
from wind_up.models import PlotConfig, WindUpConfig
from wind_up.northing_utils import add_ok_yaw_col
from wind_up.plots.northing_plots import plot_and_print_northing_error, plot_northing_changepoint, plot_northing_error

logger = logging.getLogger(__name__)


def add_rolling_northing_error(wf_df: pd.DataFrame, *, north_ref_wd_col: str, timebase_s: int) -> pd.DataFrame:
    wf_df = wf_df.copy()
    wf_df["apparent_northing_error"] = circ_diff(wf_df["YawAngleMean"], wf_df[north_ref_wd_col])
    ws_ll = 6
    wf_df.loc[~(wf_df["WindSpeedMean"] >= ws_ll), "apparent_northing_error"] = np.nan
    rolling_days = 20
    rows_per_day = 24 * 3600 / timebase_s
    wf_df["rolling_northing_error"] = (
        wf_df["apparent_northing_error"]
        .rolling(window=round(rolling_days * rows_per_day), min_periods=round(rolling_days * rows_per_day // 3))
        .median()
    )
    return wf_df


def calc_max_abs_north_errs(wf_df: pd.DataFrame, *, north_ref_wd_col: str, timebase_s: int) -> pd.DataFrame:
    wf_df = add_rolling_northing_error(wf_df.copy(), north_ref_wd_col=north_ref_wd_col, timebase_s=timebase_s)
    return wf_df.groupby("TurbineName", observed=True)["rolling_northing_error"].agg(lambda x: x.abs().max())


def calc_northed_col_name(north_ref_wd_col: str) -> str:
    return f"yawdir_northed_to_{north_ref_wd_col}"


def apply_northing_corrections(
    wf_df: pd.DataFrame,
    *,
    cfg: WindUpConfig,
    north_ref_wd_col: str,
    plot_cfg: PlotConfig | None,
    wf_north_table: pd.DataFrame | None = None,
) -> pd.DataFrame:
    wf_df = wf_df.copy()
    northed_col = calc_northed_col_name(north_ref_wd_col)

    if plot_cfg is not None:
        plot_and_print_northing_error(
            add_rolling_northing_error(wf_df, north_ref_wd_col=north_ref_wd_col, timebase_s=cfg.timebase_s),
            cfg=cfg,
            abs_north_errs=calc_max_abs_north_errs(wf_df, north_ref_wd_col=north_ref_wd_col, timebase_s=cfg.timebase_s),
            title_end=f"vs {north_ref_wd_col} before northing",
            plot_cfg=plot_cfg,
        )

    if wf_north_table is None:
        wf_north_table = pd.DataFrame(
            data=cfg.northing_corrections_utc,
            columns=["TurbineName", TIMESTAMP_COL, "north_offset"],
        )

    len_corrs = len(wf_north_table)
    if len_corrs == 0:
        if plot_cfg is not None:
            logger.info("no northing corrections to apply")
    else:
        for nc in wf_north_table.sort_values(by=["TurbineName", TIMESTAMP_COL]).itertuples():
            northing_turbine = nc.TurbineName
            northing_datetime_utc = getattr(nc, TIMESTAMP_COL)
            northing_offset = nc.north_offset
            df_idx = pd.IndexSlice[northing_turbine, northing_datetime_utc:]
            wf_df.loc[df_idx, northed_col] = (wf_df.loc[df_idx, RAW_YAWDIR_COL] + northing_offset) % 360
            if plot_cfg is not None:
                plot_northing_changepoint(
                    wf_df,
                    northing_turbine=northing_turbine,
                    northed_col=northed_col,
                    north_ref_wd_col=north_ref_wd_col,
                    northing_datetime_utc=northing_datetime_utc,
                    cfg=cfg,
                    plot_cfg=plot_cfg,
                )
            if "YawAngleMin" in wf_df.columns:
                wf_df.loc[df_idx, "YawAngleMin"] = pd.NA
            if "YawAngleMax" in wf_df.columns:
                wf_df.loc[df_idx, "YawAngleMax"] = pd.NA
        wf_df["YawAngleMean"] = wf_df[northed_col]
        if plot_cfg is not None:
            logger.info(f"applied {len_corrs} northing corrections")
            plot_and_print_northing_error(
                add_rolling_northing_error(wf_df, north_ref_wd_col=north_ref_wd_col, timebase_s=cfg.timebase_s),
                cfg=cfg,
                abs_north_errs=calc_max_abs_north_errs(
                    wf_df, north_ref_wd_col=north_ref_wd_col, timebase_s=cfg.timebase_s
                ),
                title_end=f"vs {north_ref_wd_col} after northing",
                plot_cfg=plot_cfg,
            )
    return wf_df


def format_wtg_df_like_wf_df(wtg_df: pd.DataFrame, *, wtg_name: str) -> pd.DataFrame:
    wtg_df = wtg_df.copy()
    wtg_df["TurbineName"] = wtg_name
    return wtg_df.set_index(["TurbineName", wtg_df.index])


def check_wtg_northing(
    wtg_df: pd.DataFrame,
    *,
    wtg_name: str,
    north_ref_wd_col: str,
    timebase_s: int,
    plot_cfg: PlotConfig | None,
) -> float:
    wtg_wf_df = format_wtg_df_like_wf_df(wtg_df, wtg_name=wtg_name)
    max_northing_error = calc_max_abs_north_errs(
        wtg_wf_df, north_ref_wd_col=north_ref_wd_col, timebase_s=timebase_s
    ).squeeze()
    if plot_cfg is not None:
        plot_northing_error(
            wf_df=add_rolling_northing_error(wtg_wf_df, north_ref_wd_col=north_ref_wd_col, timebase_s=timebase_s),
            title_end=f"{wtg_name} vs {north_ref_wd_col}",
            plot_cfg=plot_cfg,
            sub_dir=wtg_name,
        )
    return max_northing_error


def calc_wf_yawdir_df(
    wf_df: pd.DataFrame,
    *,
    best_yaw_dir_col: str,
    min_num_turbines: int,
    cfg: WindUpConfig,
) -> pd.DataFrame:
    wf_df = wf_df.copy()
    wf_df = add_ok_yaw_col(
        wf_df,
        new_col_name="ok_for_yawdir",
        wd_col=best_yaw_dir_col,
        rated_power=cfg.get_max_rated_power(),
        timebase_s=cfg.timebase_s,
    )
    wf_df = wf_df.loc[wf_df["ok_for_yawdir"]].dropna(subset=[best_yaw_dir_col])
    wf_yawdir_df = wf_df.groupby(TIMESTAMP_COL).agg(
        wf_yawdir=pd.NamedAgg(
            column=best_yaw_dir_col,
            aggfunc=lambda x: (
                ((x - circmean(x, low=0, high=360, nan_policy="omit") + 180) % 360).median(skipna=True)
                + circmean(x, low=0, high=360, nan_policy="omit")
                - 180
            )
            % 360,
        ),
        num_turbines=pd.NamedAgg(column=best_yaw_dir_col, aggfunc=lambda x: x.count()),
    )
    wf_yawdir_df = wf_yawdir_df[wf_yawdir_df["num_turbines"] >= min_num_turbines]
    return wf_yawdir_df.dropna()


def add_wf_yawdir(wf_df: pd.DataFrame, *, cfg: WindUpConfig) -> pd.DataFrame:
    wf_yawdir_df = calc_wf_yawdir_df(
        wf_df,
        best_yaw_dir_col=calc_northed_col_name(REANALYSIS_WD_COL),
        min_num_turbines=3,
        cfg=cfg,
    )

    wf_df_with_wf_yawdir = wf_df.drop(columns=WINDFARM_YAWDIR_COL, errors="ignore").merge(
        wf_yawdir_df[[WINDFARM_YAWDIR_COL]],
        how="left",
        left_index=True,
        right_index=True,
    )
    wf_df_with_wf_yawdir[WINDFARM_YAWDIR_COL] = wf_df_with_wf_yawdir[WINDFARM_YAWDIR_COL].fillna(
        wf_df_with_wf_yawdir[REANALYSIS_WD_COL],
    )
    return wf_df_with_wf_yawdir



import logging

import numpy as np
import pandas as pd

from wind_up.math_funcs import circ_diff
from wind_up.models import PlotConfig, WindUpConfig
from wind_up.plots.detrend_plots import (
    plot_check_applied_detrend,
    plot_detrend_ws_scatter,
    plot_detrend_wsratio_v_dir_scen,
)
from wind_up.result_manager import result_manager
from wind_up.waking_state import get_iec_upwind_turbines, lat_long_is_valid, list_wtgs_offline_in_scen

logger = logging.getLogger(__name__)


def calc_wsratio_v_wd(
    *,
    detrend_df: pd.DataFrame,
    test_ws_col: str,
    ref_ws_col: str,
    ref_wd_col: str,
    min_hours: int,
    timebase_s: int,
    dir_bin_width: float = 10.0,
) -> pd.DataFrame:
    detrend_df = detrend_df.dropna(subset=[test_ws_col, ref_ws_col, ref_wd_col]).copy()

    # IEC says only use 4-16 m/s
    test_ws_ll = 4
    test_ws_ul = 16
    ref_ws_ll = test_ws_ll * detrend_df[ref_ws_col].mean() / detrend_df[test_ws_col].mean()
    ref_ws_ul = test_ws_ul * detrend_df[ref_ws_col].mean() / detrend_df[test_ws_col].mean()
    detrend_df = detrend_df[(detrend_df[test_ws_col] >= test_ws_ll) & (detrend_df[test_ws_col] < test_ws_ul)]
    detrend_df = detrend_df[(detrend_df[ref_ws_col] >= ref_ws_ll) & (detrend_df[ref_ws_col] < ref_ws_ul)]

    directions = []
    hours = []
    test_rf_ws_roms = []
    for d in list(range(0, 360, 1)):
        detrend_df["circ_diff_to_d"] = circ_diff(detrend_df[ref_wd_col], d)
        detrend_df["within_dir_bin"] = detrend_df["circ_diff_to_d"].abs() < dir_bin_width / 2
        subsector_df = detrend_df[detrend_df["within_dir_bin"]].copy()
        if len(subsector_df) > 0:
            directions.append(d)
            rows_per_hour = 3600 / timebase_s
            hours.append(len(subsector_df) / rows_per_hour)
            # 61400-12-1 requires >=24h data, >=6h above 8m/s, >= below 8m/s
            min_count = min_hours * rows_per_hour
            accept_sector = len(subsector_df) >= min_count
            iec_ws_threshold = 8
            accept_sector = accept_sector and ((subsector_df[test_ws_col] < iec_ws_threshold).sum() >= (min_count / 4))
            accept_sector = accept_sector and ((subsector_df[test_ws_col] >= iec_ws_threshold).sum() >= (min_count / 4))
            if accept_sector:
                rom = subsector_df[test_ws_col].mean() / subsector_df[ref_ws_col].mean()
                test_rf_ws_roms.append(rom)
            else:
                test_rf_ws_roms.append(np.nan)

    return pd.DataFrame(
        {
            "direction": directions,
            "hours": hours,
            "ws_rom": test_rf_ws_roms,
        },
    )


def apply_wsratio_v_wd(
    p_df: pd.DataFrame,
    wsratio_v_dir: pd.DataFrame,
    *,
    ref_ws_col: str,
    ref_wd_col: str,
) -> pd.DataFrame:
    p_df = p_df.copy()
    wsratio_v_dir = wsratio_v_dir.copy()
    wsratio_v_dir["direction"] = wsratio_v_dir.index
    # add a 360 row to wsratio_v_dir
    rows_w_0_dir = wsratio_v_dir[wsratio_v_dir["direction"] == 0].copy()
    rows_w_0_dir["direction"] = 360
    wsratio_v_dir = pd.concat([wsratio_v_dir, rows_w_0_dir]).reset_index(drop=True)

    p_df["ws_rom"] = np.interp(p_df[ref_wd_col].to_numpy(), wsratio_v_dir["direction"], wsratio_v_dir["ws_rom"])

    p_df["ref_ws_detrended"] = p_df[ref_ws_col] * p_df["ws_rom"]

    return p_df


def remove_bad_detrend_results(
    wsratio_v_dir_scen: pd.DataFrame,
    test_name: str,
    ref_name: str,
    ref_lat: float,
    ref_long: float,
    cfg: WindUpConfig,
) -> pd.DataFrame:
    try:
        none_offline_df = wsratio_v_dir_scen.dropna(subset="ws_rom").loc["none offline"]
    except KeyError:
        result_manager.warning("cannot remove_bad_detrend_results, no 'none offline' rows with ws_rom defined")
        return wsratio_v_dir_scen

    try:
        test_wtg = next(x for x in cfg.asset.wtgs if x.name == test_name)
    except StopIteration as exc:
        msg = f"{test_name} not found in cfg.asset.wtgs"
        raise ValueError(msg) from exc

    remove_count = 0
    for wd in none_offline_df.index.unique():
        test_upwind_wtgs = get_iec_upwind_turbines(
            latlongs=test_wtg.get_latlongs(),
            wind_direction=wd,
            cfg=cfg,
            object_name=test_name,
        )
        if lat_long_is_valid(ref_lat, ref_long):
            ref_upwind_wtgs = get_iec_upwind_turbines(
                latlongs=[(ref_lat, ref_long)],
                wind_direction=wd,
                cfg=cfg,
                object_name=ref_name,
            )
        else:
            ref_upwind_wtgs = []

        scens_to_check = [
            x
            for x in wsratio_v_dir_scen.dropna(subset="ws_rom")
            .loc[pd.IndexSlice[:, wd], :]
            .index.unique(level="waking_scenario")
            if x != "none offline"
        ]
        for scen in scens_to_check:
            upwind_wtgs_offline = list_wtgs_offline_in_scen(scen)
            test_upwind_wtgs_offline = sorted(set(test_upwind_wtgs) & set(upwind_wtgs_offline))
            ref_upwind_wtgs_offline = sorted(set(ref_upwind_wtgs) & set(upwind_wtgs_offline))
            if (len(test_upwind_wtgs_offline) + len(ref_upwind_wtgs_offline) > 0) and (
                len(test_upwind_wtgs_offline) == 0 or len(ref_upwind_wtgs_offline) == 0
            ):
                sign = 1 if len(ref_upwind_wtgs_offline) == 0 else -1
                scen_ws_rom = wsratio_v_dir_scen.loc[pd.IndexSlice[scen, wd], "ws_rom"]
                none_offline_ws_rom = none_offline_df.loc[wd, "ws_rom"]
                abs_tol = 1e-6
                if sign * scen_ws_rom < (sign * none_offline_ws_rom - abs_tol):
                    wsratio_v_dir_scen.loc[pd.IndexSlice[scen, wd], "ws_rom"] = np.nan
                    remove_count += 1

    logger.info(f"removed {remove_count} bad detrend results")
    return wsratio_v_dir_scen


def calc_wsratio_v_wd_scen(
    *,
    test_name: str,
    ref_name: str,
    ref_lat: float,
    ref_long: float,
    detrend_df: pd.DataFrame,
    test_ws_col: str,
    ref_ws_col: str,
    ref_wd_col: str,
    cfg: WindUpConfig,
    plot_cfg: PlotConfig | None,
) -> pd.DataFrame:
    count_by_scen = detrend_df.dropna(subset=[test_ws_col, ref_ws_col, ref_wd_col])["waking_scenario"].value_counts()
    rows_per_hour = 3600 / cfg.timebase_s
    scens_to_detrend = count_by_scen[count_by_scen > (cfg.detrend_min_hours * rows_per_hour)]
    scens_to_detrend_list = [x for x in scens_to_detrend.index if x != "unknown"]

    wsratio_v_dir_scen = pd.DataFrame()
    if len(scens_to_detrend_list) == 0:
        result_manager.warning("no scenarios with enough data to detrend")
        return pd.DataFrame(columns=["hours", "ws_rom"])

    for scen in scens_to_detrend_list:
        scen_df = detrend_df[detrend_df["waking_scenario"] == scen]
        scen_df = scen_df.dropna(subset=[test_ws_col, ref_ws_col, ref_wd_col])
        wsratio_v_dir = calc_wsratio_v_wd(
            detrend_df=scen_df,
            test_ws_col=test_ws_col,
            ref_ws_col=ref_ws_col,
            ref_wd_col=ref_wd_col,
            min_hours=cfg.detrend_min_hours,
            timebase_s=cfg.timebase_s,
        )
        wsratio_v_dir["waking_scenario"] = scen
        wsratio_v_dir_scen = pd.concat([wsratio_v_dir_scen, wsratio_v_dir])

    wsratio_v_dir_scen = wsratio_v_dir_scen.set_index(["waking_scenario", "direction"])

    wsratio_v_dir_scen = remove_bad_detrend_results(
        wsratio_v_dir_scen,
        test_name=test_name,
        ref_name=ref_name,
        ref_lat=ref_lat,
        ref_long=ref_long,
        cfg=cfg,
    )

    if plot_cfg is not None:
        plot_detrend_ws_scatter(
            detrend_df=detrend_df,
            test_name=test_name,
            ref_name=ref_name,
            test_ws_col=test_ws_col,
            ref_ws_col=ref_ws_col,
            plot_cfg=plot_cfg,
        )

        plot_detrend_wsratio_v_dir_scen(
            wsratio_v_dir_scen=wsratio_v_dir_scen,
            test_name=test_name,
            ref_name=ref_name,
            ref_wd_col=ref_wd_col,
            plot_cfg=plot_cfg,
        )
    return wsratio_v_dir_scen


def apply_wsratio_v_wd_scen(
    p_df: pd.DataFrame,
    wsratio_v_dir_scen: pd.DataFrame,
    *,
    ref_ws_col: str,
    ref_wd_col: str,
) -> pd.DataFrame:
    scen_list = list(
        set(p_df.dropna(subset=[ref_ws_col, ref_wd_col])["waking_scenario"].unique())
        & set(wsratio_v_dir_scen.dropna(subset="ws_rom").index.unique(level="waking_scenario")),
    )
    all_scens_df = pd.DataFrame()
    for scen in scen_list:
        scen_df = p_df[p_df["waking_scenario"] == scen]
        scen_df = scen_df.dropna(subset=[ref_ws_col, ref_wd_col])
        wsratio_v_dir = wsratio_v_dir_scen.loc[scen].dropna(subset="ws_rom")
        scen_df = apply_wsratio_v_wd(scen_df, wsratio_v_dir, ref_ws_col=ref_ws_col, ref_wd_col=ref_wd_col)
        all_scens_df = pd.concat([all_scens_df, scen_df])

    columns_to_add = ["ws_rom", "ref_ws_detrended"]

    try:
        result_df = p_df.merge(all_scens_df[columns_to_add], how="left", left_index=True, right_index=True)
    except KeyError:
        result_manager.warning("no rows in p_df to merge with all_scens_df")
        result_df = p_df.copy()
        result_df[columns_to_add] = np.nan
    # print count of distinct scenario - directions
    count_detrend_applied_df = result_df.dropna(subset=[ref_ws_col, ref_wd_col]).copy()
    count_detrend_applied_df["rounded_wd"] = count_detrend_applied_df[ref_wd_col].round(0).mod(360)
    count_detrend_applied_df["scen_wd"] = (
        count_detrend_applied_df["waking_scenario"] + "_" + count_detrend_applied_df["rounded_wd"].astype(str)
    )
    logger.info(f"detrend applied to {len(count_detrend_applied_df['scen_wd'].unique())} scenario - directions")

    return result_df


def check_applied_detrend(
    *,
    test_name: str,
    ref_name: str,
    ref_lat: float,
    ref_long: float,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    test_ws_col: str,
    ref_ws_col: str,
    detrend_ws_col: str,
    ref_wd_col: str,
    cfg: WindUpConfig,
    plot_cfg: PlotConfig | None,
) -> tuple[float, float]:
    pre_df = pre_df.dropna(subset=[test_ws_col, ref_ws_col, detrend_ws_col, ref_wd_col]).copy()
    post_df = post_df.dropna(subset=[test_ws_col, ref_ws_col, detrend_ws_col, ref_wd_col]).copy()

    # confirm pre_df correlation has improved thanks to detrend
    pre_r2_before_detrend = pre_df[test_ws_col].corr(pre_df[ref_ws_col]) ** 2
    pre_r2_after_detrend = pre_df[test_ws_col].corr(pre_df[detrend_ws_col]) ** 2
    pre_r2_improvement = pre_r2_after_detrend - pre_r2_before_detrend
    if pre_r2_improvement >= 0:
        logger.info(
            f"detrend improved pre_df ws r2 by {pre_r2_improvement:.2f} "
            f"({pre_r2_before_detrend:.2f} to {pre_r2_after_detrend:.2f})",
        )
    else:
        msg = f"pre_r2_after_detrend < pre_r2_before_detrend, {pre_r2_after_detrend} < {pre_r2_before_detrend}"
        result_manager.warning(msg)

    # print post_df corr change
    post_r2_before_detrend = post_df[test_ws_col].corr(post_df[ref_ws_col]) ** 2
    post_r2_after_detrend = post_df[test_ws_col].corr(post_df[detrend_ws_col]) ** 2
    post_r2_improvement = post_r2_after_detrend - post_r2_before_detrend
    if post_r2_improvement >= 0:
        logger.info(
            f"detrend improved post_df ws r2 by {post_r2_improvement:.2f} "
            f"({post_r2_before_detrend:.2f} to {post_r2_after_detrend:.2f})",
        )
    else:
        result_manager.warning(
            f"post_r2_after_detrend < post_r2_before_detrend," f" {post_r2_after_detrend} < {post_r2_before_detrend}"
        )

    if plot_cfg is not None:
        pre_wsratio_v_dir_scen = calc_wsratio_v_wd_scen(
            test_name=test_name,
            ref_name=ref_name,
            ref_lat=ref_lat,
            ref_long=ref_long,
            detrend_df=pre_df,
            test_ws_col=test_ws_col,
            ref_ws_col=detrend_ws_col,
            ref_wd_col=ref_wd_col,
            cfg=cfg,
            plot_cfg=None,
        )
        post_wsratio_v_dir_scen = calc_wsratio_v_wd_scen(
            test_name=test_name,
            ref_name=ref_name,
            ref_lat=ref_lat,
            ref_long=ref_long,
            detrend_df=post_df,
            test_ws_col=test_ws_col,
            ref_ws_col=detrend_ws_col,
            ref_wd_col=ref_wd_col,
            cfg=cfg,
            plot_cfg=None,
        )

        plot_check_applied_detrend(
            pre_df=pre_df,
            post_df=post_df,
            test_name=test_name,
            ref_name=ref_name,
            test_ws_col=test_ws_col,
            ref_ws_col=ref_ws_col,
            detrend_ws_col=detrend_ws_col,
            ref_wd_col=ref_wd_col,
            pre_wsratio_v_dir_scen=pre_wsratio_v_dir_scen,
            post_wsratio_v_dir_scen=post_wsratio_v_dir_scen,
            plot_cfg=plot_cfg,
        )

    return pre_r2_improvement, post_r2_improvement



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from wind_up.constants import RAW_WINDSPEED_COL, SCATTER_ALPHA, SCATTER_MARKERSCALE, SCATTER_S, DataColumns
from wind_up.models import PlotConfig
from wind_up.result_manager import result_manager


def plot_pre_post_binned_power_curves(
    *,
    test_name: str,
    ref_name: str,
    pp_df: pd.DataFrame,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure()
    plt.errorbar(pp_df["ws_mean_pre"], pp_df["pw_mean_pre_raw"], yerr=pp_df["pw_sem_pre"], label="pre raw", marker=".")
    plt.errorbar(
        pp_df["ws_mean_post"],
        pp_df["pw_mean_post_raw"],
        yerr=pp_df["pw_sem_post"],
        label="post raw",
        marker=".",
    )
    plt.legend()
    plt.grid()
    plot_title = f"test={test_name} ref={ref_name} raw power curve"
    plt.title(plot_title)
    plt.ylabel("mean power [kW]")
    plt.xlabel("mean wind speed [m/s]")
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.errorbar(pp_df["bin_mid"], pp_df["pw_at_mid_pre"], yerr=pp_df["pw_sem_at_mid_pre"], label="pre", marker=".")
    plt.errorbar(
        pp_df["bin_mid"],
        pp_df["pw_at_mid_post"],
        yerr=pp_df["pw_sem_at_mid_post"],
        label="post",
        marker=".",
    )
    plt.legend()
    plt.grid()
    plot_title = f"test={test_name} ref={ref_name} cooked power curve"
    plt.title(plot_title)
    plt.ylabel("power at bin mid [kW]")
    plt.xlabel("wind speed bin mid [m/s]")
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.errorbar(
        pp_df["bin_mid"],
        pp_df["pw_at_mid_expected"],
        yerr=pp_df["pw_sem_at_mid_expected"],
        label="baseline",
        marker=".",
    )
    plt.errorbar(
        pp_df["bin_mid"],
        pp_df["pw_at_mid_post"],
        yerr=pp_df["pw_sem_at_mid_post"],
        label="post",
        marker=".",
    )
    plt.legend()
    plt.grid()
    plot_title = f"test={test_name} ref={ref_name} power curves for uplift"
    plt.title(plot_title)
    plt.ylabel("power at bin mid [kW]")
    plt.xlabel("wind speed bin mid [m/s]")
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.errorbar(
        pp_df["bin_mid"],
        pp_df["relative_cp_baseline"],
        yerr=pp_df["relative_cp_sem_at_mid_expected"],
        label="baseline",
        marker=".",
    )
    plt.errorbar(
        pp_df["bin_mid"],
        pp_df["relative_cp_post"],
        yerr=pp_df["relative_cp_sem_at_mid_post"],
        label="post",
        marker=".",
    )
    plt.legend()
    plt.grid()
    plot_title = f"test={test_name} ref={ref_name} relative Cp"
    plt.title(plot_title)
    plt.ylabel("Cp relative to baseline max [-]")
    plt.xlabel("wind speed bin mid [m/s]")
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()


def plot_pre_post_power_curves(
    *,
    test_name: str,
    ref_name: str,
    pp_df: pd.DataFrame,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    ws_col: str,
    pw_col: str,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure()
    plt.scatter(pre_df[ws_col], pre_df[pw_col], s=SCATTER_S, alpha=SCATTER_ALPHA, label="pre upgrade")
    plt.scatter(post_df[ws_col], post_df[pw_col], s=SCATTER_S, alpha=SCATTER_ALPHA, label="post upgrade")
    plt.legend(loc="best", markerscale=SCATTER_MARKERSCALE)
    plt.grid()
    plot_title = f"test={test_name} ref={ref_name} power curve data"
    plt.title(plot_title)
    plt.ylabel(f"{pw_col} [kW]")
    plt.xlabel(f"{ws_col} [m/s]")
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.plot(pp_df["bin_mid"], pp_df["hours_pre"] / pp_df["hours_pre"].sum(), marker="s", label="pre")
    plt.plot(pp_df["bin_mid"], pp_df["hours_post"] / pp_df["hours_post"].sum(), marker="s", label="post")
    plt.plot(pp_df["bin_mid"], pp_df["f"], marker="s", label="distribution used for uplift")
    plt.legend()
    plt.grid()
    plot_title = f"test={test_name} ref={ref_name} relative frequency"
    plt.title(plot_title)
    plt.ylabel("fraction of time [-]")
    plt.xlabel("bin mid [m/s]")
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()

    plot_pre_post_binned_power_curves(
        test_name=test_name,
        ref_name=ref_name,
        pp_df=pp_df,
        plot_cfg=plot_cfg,
    )


def plot_pre_post_condition_histogram(
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    *,
    test_name: str,
    ref_name: str,
    rows_per_hour: float,
    col: str,
    bin_width: float,
    plot_cfg: PlotConfig,
    first_bin_start: float | None = None,
    last_bin_end: float | None = None,
    x_ticks: np.ndarray | None = None,
) -> None:
    if col not in pre_df.columns:
        msg = f"plot_pre_post_condition_histogram ref_name={ref_name} pre_df missing required column {col}"
        result_manager.warning(msg)
        return
    if col not in post_df.columns:
        msg = f"plot_pre_post_condition_histogram ref_name={ref_name} post_df missing required column {col}"
        result_manager.warning(msg)
        return
    if first_bin_start is None:
        first_bin_start = round(min(pre_df[col].min(), post_df[col].min()) - bin_width / 2)
    if last_bin_end is None:
        last_bin_end = max(pre_df[col].max(), post_df[col].max())
    bins = list(
        np.arange(
            first_bin_start,
            last_bin_end + bin_width / 2,
            bin_width,
        )
    )
    plt.figure()
    plt.hist(
        pre_df[col],
        weights=[1 / rows_per_hour] * len(pre_df[col]),
        bins=bins,
        label="pre",
    )
    plt.hist(
        post_df[col],
        weights=[1 / rows_per_hour] * len(post_df[col]),
        bins=bins,
        alpha=0.5,
        label="post",
    )
    plot_title = f"{test_name} {ref_name} {col} coverage"
    plt.title(plot_title)
    if x_ticks is not None:
        plt.xticks(x_ticks)
    plt.xlabel(col)
    plt.ylabel("hours")
    plt.legend()
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()


def plot_pre_post_conditions(
    *,
    test_name: str,
    ref_name: str,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    ws_col: str,
    wd_col: str,
    timebase_s: int,
    plot_cfg: PlotConfig,
) -> None:
    rows_per_hour = 3600 / timebase_s

    hod_width = 1
    plt.figure()
    plt.hist(
        pre_df.index.hour,
        weights=[1 / rows_per_hour] * len(pre_df.index.hour),
        bins=list(np.arange(-hod_width / 2, 24 + hod_width / 2, hod_width)),
        label="pre",
    )
    plt.hist(
        post_df.index.hour,
        weights=[1 / rows_per_hour] * len(post_df.index.hour),
        bins=list(np.arange(-hod_width / 2, 24 + hod_width / 2, hod_width)),
        alpha=0.5,
        label="post",
    )
    plot_title = f"{test_name} {ref_name} hour of day coverage"
    plt.title(plot_title)
    plt.xticks(np.arange(0, 25, 4))
    plt.ylabel("hours")
    plt.xlabel("hour of day")
    plt.legend()
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()

    moy_width = 1
    plt.figure()
    plt.hist(
        pre_df.index.month,
        weights=[1 / rows_per_hour] * len(pre_df.index.month),
        bins=list(np.arange(0.5, 12 + moy_width, moy_width)),
        label="pre",
    )
    plt.hist(
        post_df.index.month,
        weights=[1 / rows_per_hour] * len(post_df.index.month),
        bins=list(np.arange(0.5, 12 + moy_width, moy_width)),
        alpha=0.5,
        label="post",
    )
    plot_title = f"{test_name} {ref_name} month of year coverage"
    plt.title(plot_title)
    plt.xticks(np.arange(1, 13, 1))
    plt.ylabel("hours")
    plt.xlabel("month of year")
    plt.legend()
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()

    wd_width = 30
    plot_pre_post_condition_histogram(
        pre_df,
        post_df,
        test_name=test_name,
        ref_name=ref_name,
        rows_per_hour=rows_per_hour,
        col=wd_col,
        bin_width=wd_width,
        plot_cfg=plot_cfg,
        first_bin_start=0,
        last_bin_end=360,
        x_ticks=np.arange(wd_width / 2, 360 + wd_width / 2, wd_width),
    )
    plot_pre_post_condition_histogram(
        pre_df,
        post_df,
        test_name=test_name,
        ref_name=ref_name,
        rows_per_hour=rows_per_hour,
        col="ref_" + DataColumns.ambient_temp,
        bin_width=2,
        plot_cfg=plot_cfg,
    )
    plot_pre_post_condition_histogram(
        pre_df,
        post_df,
        test_name=test_name,
        ref_name=ref_name,
        rows_per_hour=rows_per_hour,
        col=ws_col,
        bin_width=1,
        plot_cfg=plot_cfg,
        first_bin_start=0,
    )


def plot_pre_post_pp_analysis(
    *,
    test_name: str,
    ref_name: str,
    pp_df: pd.DataFrame,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    ws_col: str,
    pw_col: str,
    wd_col: str,
    timebase_s: int,
    plot_cfg: PlotConfig,
    confidence_level: float,
) -> None:
    plot_pre_post_conditions(
        test_name=test_name,
        ref_name=ref_name,
        pre_df=pre_df.dropna(subset=[ws_col, pw_col]),
        post_df=post_df.dropna(subset=[ws_col, pw_col]),
        ws_col=ws_col,
        wd_col=wd_col,
        timebase_s=timebase_s,
        plot_cfg=plot_cfg,
    )

    pp_df = pp_df.copy()
    p_low = (1 - confidence_level) / 2
    p_high = 1 - p_low

    plot_pre_post_power_curves(
        test_name=test_name,
        ref_name=ref_name,
        pp_df=pp_df,
        pre_df=pre_df.dropna(subset=[ws_col, pw_col]),
        post_df=post_df.dropna(subset=[ws_col, pw_col]),
        ws_col=ws_col,
        pw_col=pw_col,
        plot_cfg=plot_cfg,
    )

    plt.figure()
    plt.plot(pp_df["bin_mid"], pp_df["uplift_kw"], color="b", marker="s")
    plt.plot(pp_df["bin_mid"], pp_df[f"uplift_p{p_low*100:.0f}_kw"], color="r", ls="--")
    plt.plot(pp_df["bin_mid"], pp_df[f"uplift_p{p_high*100:.0f}_kw"], color="r", ls="--")
    plt.grid()
    plot_title = f"test={test_name} ref={ref_name} uplift [kW] and {confidence_level*100:.0f}% CI"
    plt.title(plot_title)
    plt.ylabel("uplift [kW]")
    plt.xlabel("bin centre [m/s]")
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.plot(pp_df["bin_mid"], pp_df["uplift_relative_cp"], color="b", marker="s")
    plt.plot(pp_df["bin_mid"], pp_df[f"uplift_relative_cp_p{p_low * 100:.0f}"], color="r", ls="--")
    plt.plot(pp_df["bin_mid"], pp_df[f"uplift_relative_cp_p{p_high * 100:.0f}"], color="r", ls="--")
    plt.grid()
    plot_title = f"test={test_name} ref={ref_name} relative Cp uplift and {confidence_level * 100:.0f}% CI"
    plt.title(plot_title)
    plt.ylabel("change in Cp normalized to max Cp")
    plt.xlabel("bin centre [m/s]")
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()

    plt.figure()
    plt.plot(
        pp_df["bin_mid"],
        pp_df["uplift_kw"] * pp_df["hours_for_mwh_calc"].sum() * pp_df["f"] / 1000,
        color="b",
        marker="s",
    )
    plt.plot(
        pp_df["bin_mid"],
        pp_df[f"uplift_p{p_low*100:.0f}_kw"] * pp_df["hours_for_mwh_calc"].sum() * pp_df["f"] / 1000,
        color="r",
        ls="--",
    )
    plt.plot(
        pp_df["bin_mid"],
        pp_df[f"uplift_p{p_high*100:.0f}_kw"] * pp_df["hours_for_mwh_calc"].sum() * pp_df["f"] / 1000,
        color="r",
        ls="--",
    )
    plt.grid()
    plot_title = f"test={test_name} ref={ref_name} uplift [MWh] and {confidence_level*100:.0f}% CI"
    plt.title(plot_title)
    plt.ylabel("uplift [MWh]")
    plt.xlabel("bin centre [m/s]")
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()


def plot_pp_data_coverage(
    *,
    test_name: str,
    ref_name: str,
    pp_df: pd.DataFrame,
    test_df_pp_period: pd.DataFrame,
    ws_bin_width: float,
    timebase_s: int,
    plot_cfg: PlotConfig,
) -> None:
    rows_per_hour = 3600 / timebase_s
    ws_bin_edges = np.arange(0, test_df_pp_period["test_" + RAW_WINDSPEED_COL].max() + ws_bin_width, ws_bin_width)
    raw_df = test_df_pp_period.groupby(
        by=pd.cut(test_df_pp_period["test_" + RAW_WINDSPEED_COL], bins=ws_bin_edges, retbins=False),
        observed=False,
    ).agg(
        hours_raw=pd.NamedAgg(column="test_" + RAW_WINDSPEED_COL, aggfunc=lambda x: x.count() / rows_per_hour),
    )
    raw_df["bin_mid"] = [x.mid for x in raw_df.index]

    plot_df = pp_df.merge(raw_df, on="bin_mid", how="left")
    plot_df["data_coverage"] = (
        ((plot_df["hours_pre"].fillna(0) + plot_df["hours_post"].fillna(0)) / plot_df["hours_raw"])
        .clip(upper=1)
        .fillna(1)
    )

    plt.figure()
    plt.plot(plot_df["bin_mid"], 100 * plot_df["data_coverage"], marker=".")
    plot_title = f"{test_name} {ref_name} data coverage"
    plt.title(plot_title)
    plt.ylabel("data coverage [%]")
    plt.xlabel("bin centre [m/s]")
    plt.ylim([0, 100])
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / test_name / f"{plot_title}.png")
    plt.close()



import pandas as pd
import pytest
from pandas.testing import assert_series_equal

from wind_up.wind_funcs import calc_cp

test_calc_cp_data = [
    (500, 8.2, 1.21, 62, 0.4964778),
    (pd.Series([500.1, -30.1]), pd.Series([8.4, 8.5]), 1.21, 62, pd.Series([0.4619451, -0.0268337])),
]


@pytest.mark.parametrize(("power_kw", "ws_ms", "air_density_kgpm3", "rotor_diamter_m", "expected"), test_calc_cp_data)
def test_calc_cp(
    power_kw: float | pd.Series,
    ws_ms: float | pd.Series,
    air_density_kgpm3: float | pd.Series,
    rotor_diamter_m: float,
    expected: float | pd.Series,
) -> None:
    if isinstance(expected, pd.Series):
        assert_series_equal(calc_cp(power_kw, ws_ms, air_density_kgpm3, rotor_diamter_m), (expected))
    else:
        assert calc_cp(power_kw, ws_ms, air_density_kgpm3, rotor_diamter_m) == pytest.approx(expected)

    cp = calc_cp(500, 8.2, 1.21, 62)
    assert cp == pytest.approx(0.4964778)



import contextlib
import logging
import math

import numpy as np
import pandas as pd
from scipy.stats import norm, t
from tqdm.auto import tqdm

from wind_up.models import PlotConfig, Turbine, WindUpConfig
from wind_up.plots.pp_analysis_plots import plot_pp_data_coverage, plot_pre_post_pp_analysis
from wind_up.result_manager import result_manager

logger = logging.getLogger(__name__)


def pp_raw_df(
    pre_or_post_df: pd.DataFrame,
    pre_or_post: str,
    *,
    ws_col: str,
    ws_bin_edges: np.ndarray,
    pw_col: str,
    timebase_s: int,
) -> pd.DataFrame:
    pp_df = (
        pre_or_post_df.dropna(subset=[pw_col, ws_col])
        .groupby(
            by=pd.cut(pre_or_post_df[ws_col], bins=ws_bin_edges, retbins=False),
            observed=False,
        )
        .agg(
            count=pd.NamedAgg(column=pw_col, aggfunc=lambda x: len(x)),
            ws_mean=pd.NamedAgg(column=ws_col, aggfunc=lambda x: x.mean()),
            ws_std=pd.NamedAgg(column=ws_col, aggfunc=lambda x: x.std()),
            pw_mean=pd.NamedAgg(column=pw_col, aggfunc=lambda x: x.mean()),
            pw_std=pd.NamedAgg(column=pw_col, aggfunc=lambda x: x.std()),
        )
    )
    pp_df["ws_std"] = pp_df["ws_std"].fillna(0)
    pp_df["pw_std"] = pp_df["pw_std"].fillna(0)
    rows_per_hour = 3600 / timebase_s
    pp_df["hours"] = pp_df["count"] / rows_per_hour
    pp_df["ws_sem"] = pp_df["ws_std"] / np.sqrt(pp_df["count"].clip(lower=1))
    pp_df["pw_sem"] = pp_df["pw_std"] / np.sqrt(pp_df["count"].clip(lower=1))
    pp_df.columns = [x + f"_{pre_or_post}" for x in pp_df.columns]
    pp_df["bin_left"] = [x.left for x in pp_df.index]
    pp_df["bin_mid"] = [x.mid for x in pp_df.index]
    pp_df["bin_right"] = [x.right for x in pp_df.index]
    pp_df["bin_closed_right"] = [x.closed_right for x in pp_df.index]
    pp_df = pp_df.set_index("bin_mid", drop=False, verify_integrity=True)
    pp_df.index.name = f"{ws_col}_bin_mid"
    return pp_df


def cook_pp(pp_df: pd.DataFrame, pre_or_post: str, ws_bin_width: float, rated_power: float) -> pd.DataFrame:
    pp_df = pp_df.copy()

    valid_col = f"{pre_or_post}_valid"
    raw_pw_col = f"pw_mean_{pre_or_post}_raw"
    raw_hours_col = f"hours_{pre_or_post}_raw"
    pw_col = f"pw_mean_{pre_or_post}"
    pw_sem_col = f"pw_sem_{pre_or_post}"
    hours_col = f"hours_{pre_or_post}"
    ws_col = f"ws_mean_{pre_or_post}"
    pw_at_mid_col = f"pw_at_mid_{pre_or_post}"
    pw_sem_at_mid_col = f"pw_sem_at_mid_{pre_or_post}"

    pp_df[raw_pw_col] = pp_df[pw_col]
    pp_df[raw_hours_col] = pp_df[hours_col]

    # IEC minimum data count method
    hrs_per_mps = 1
    pp_df[valid_col] = pp_df[f"hours_{pre_or_post}"] > (ws_bin_width * hrs_per_mps)
    pp_df.loc[~pp_df[valid_col], [pw_col, hours_col, ws_col, pw_sem_col]] = np.nan
    pp_df[hours_col] = pp_df[hours_col].fillna(0)
    pp_df[ws_col] = pp_df[ws_col].fillna(pp_df["bin_mid"])

    pp_df[pw_col] = pp_df[pw_col].clip(lower=0, upper=rated_power)

    # data which would have been at rated can be gap filled
    rated_ws = pp_df.loc[pp_df[raw_pw_col] >= rated_power * 0.995, "bin_mid"].min() + 1
    pp_df.loc[(pp_df["bin_mid"] >= rated_ws) & pp_df[pw_col].isna(), pw_col] = rated_power
    pp_df[pw_sem_col] = pp_df[pw_sem_col].ffill()

    # missing data at low wind speed can be filled with 0
    pp_df.loc[pp_df.index[pp_df[pw_col].isna().cummin()], [pw_col, pw_sem_col]] = 0

    # revisit in future, an alternative is filling in with SCADA power curve
    pp_df[pw_col] = pp_df[pw_col].interpolate()

    # interpolate power and ci to bin mid
    pp_df[pw_at_mid_col] = np.interp(pp_df["bin_mid"], pp_df[ws_col], pp_df[pw_col])
    pp_df[pw_at_mid_col] = pp_df[pw_at_mid_col].clip(lower=0, upper=rated_power)
    pp_df[pw_sem_at_mid_col] = pp_df[pw_sem_col] / pp_df[pw_col] * pp_df[pw_at_mid_col]
    pp_df[pw_sem_at_mid_col] = pp_df[pw_sem_at_mid_col].fillna(0)
    pp_df[pw_sem_at_mid_col] = pp_df[pw_sem_at_mid_col].clip(lower=0, upper=pp_df[pw_sem_col])

    if pp_df[[col for col in pp_df.columns if col is not raw_pw_col]].isna().any().any():
        msg = "pp_df has missing values"
        result_manager.warning(msg)

    return pp_df


def add_uplift_cols_to_pp_df(pp_df: pd.DataFrame, *, p_low: float, p_high: float, t_values: np.ndarray) -> pd.DataFrame:
    new_pp_df = pp_df.copy()
    # calculations needed for uplift vs wind speed plot
    new_pp_df["uplift_kw"] = new_pp_df["pw_at_mid_post"] - new_pp_df["pw_at_mid_expected"]
    new_pp_df["uplift_kw_se"] = np.sqrt(new_pp_df["pw_sem_at_mid_post"] ** 2 + new_pp_df["pw_sem_at_mid_expected"] ** 2)
    new_pp_df[f"uplift_p{p_low * 100:.0f}_kw"] = new_pp_df["uplift_kw"] + new_pp_df["uplift_kw_se"] * t_values
    new_pp_df[f"uplift_p{p_high * 100:.0f}_kw"] = new_pp_df["uplift_kw"] - new_pp_df["uplift_kw_se"] * t_values

    # calculations needed for relative Cp plots
    new_pp_df["relative_cp_baseline"] = new_pp_df["pw_at_mid_expected"] / new_pp_df["bin_mid"] ** 3
    max_baseline_cp = new_pp_df["relative_cp_baseline"].max()
    new_pp_df["relative_cp_baseline"] = new_pp_df["relative_cp_baseline"] / max_baseline_cp
    new_pp_df["relative_cp_post"] = new_pp_df["pw_at_mid_post"] / new_pp_df["bin_mid"] ** 3 / max_baseline_cp
    new_pp_df["relative_cp_sem_at_mid_expected"] = (
        new_pp_df["pw_sem_at_mid_expected"] / new_pp_df["bin_mid"] ** 3 / max_baseline_cp
    )
    new_pp_df["relative_cp_sem_at_mid_post"] = (
        new_pp_df["pw_sem_at_mid_post"] / new_pp_df["bin_mid"] ** 3 / max_baseline_cp
    )
    new_pp_df["uplift_relative_cp"] = new_pp_df["relative_cp_post"] - new_pp_df["relative_cp_baseline"]
    new_pp_df["uplift_relative_cp_se"] = np.sqrt(
        new_pp_df["relative_cp_sem_at_mid_post"] ** 2 + new_pp_df["relative_cp_sem_at_mid_expected"] ** 2
    )
    new_pp_df[f"uplift_relative_cp_p{p_low * 100:.0f}"] = (
        new_pp_df["uplift_relative_cp"] + new_pp_df["uplift_relative_cp_se"] * t_values
    )
    new_pp_df[f"uplift_relative_cp_p{p_high * 100:.0f}"] = (
        new_pp_df["uplift_relative_cp"] - new_pp_df["uplift_relative_cp_se"] * t_values
    )

    return new_pp_df


def pre_post_pp_analysis(
    *,
    cfg: WindUpConfig,
    test_wtg: Turbine,
    ref_name: str,
    lt_df: pd.DataFrame | None,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    ws_col: str,
    pw_col: str,
    wd_col: str,
    plot_cfg: PlotConfig | None,
    confidence_level: float = 0.9,
    test_df: pd.DataFrame | None = None,
    reverse: bool = False,
) -> tuple[dict, pd.DataFrame]:
    wtg_for_turbine_type = test_wtg
    test_name = test_wtg.name
    if reverse:
        with contextlib.suppress(StopIteration):
            wtg_for_turbine_type = next(x for x in cfg.asset.wtgs if x.name == ref_name)
        test_name, ref_name = ref_name, test_name
        pre_df, post_df = post_df, pre_df
        ws_col = ws_col.replace("ref", "test")
        pw_col = pw_col.replace("test", "ref")
    cutout_ws = wtg_for_turbine_type.turbine_type.cutout_ws_mps
    rated_power = wtg_for_turbine_type.turbine_type.rated_power_kw

    ws_bin_edges = np.arange(0, cutout_ws + cfg.ws_bin_width, cfg.ws_bin_width)

    pre_pp_df = pp_raw_df(
        pre_df, "pre", ws_col=ws_col, ws_bin_edges=ws_bin_edges, pw_col=pw_col, timebase_s=cfg.timebase_s
    )
    post_pp_df = pp_raw_df(
        post_df, "post", ws_col=ws_col, ws_bin_edges=ws_bin_edges, pw_col=pw_col, timebase_s=cfg.timebase_s
    )

    pre_pp_df = cook_pp(pp_df=pre_pp_df, pre_or_post="pre", ws_bin_width=cfg.ws_bin_width, rated_power=rated_power)
    post_pp_df = cook_pp(pp_df=post_pp_df, pre_or_post="post", ws_bin_width=cfg.ws_bin_width, rated_power=rated_power)
    pp_df = pre_pp_df.merge(
        post_pp_df[[x for x in post_pp_df.columns if x not in pre_pp_df.columns]],
        on=pre_pp_df.index.name,
        how="inner",
    )
    pp_df["is_invalid_bin"] = (~pp_df["pre_valid"]) | (~pp_df["post_valid"])
    if lt_df is not None:
        pp_df = pp_df.merge(lt_df[["hours_per_year", "bin_mid"]], on="bin_mid", how="left")
        pp_df = pp_df.set_index("bin_mid", drop=False, verify_integrity=True)
        pp_df.index.name = f"{ws_col}_bin_mid"
        pp_df["hours_for_mwh_calc"] = pp_df["hours_per_year"].fillna(0)
    else:
        pp_df["hours_for_mwh_calc"] = pp_df["hours_post_raw"].fillna(0)
    pp_df["f"] = pp_df["hours_for_mwh_calc"] / pp_df["hours_for_mwh_calc"].sum()

    if lt_df is not None:
        pp_df_invalid = pp_df[pp_df["is_invalid_bin"]]
        mwh_invalid_bins = (
            pp_df_invalid["hours_for_mwh_calc"].sum()
            * (pp_df_invalid["f"] * pp_df_invalid["pw_at_mid_pre"]).sum()
            / 1000
        )
        pre_mwh = pp_df["hours_for_mwh_calc"].sum() * (pp_df["f"] * pp_df["pw_at_mid_pre"]).sum() / 1000
        missing_bins_unc_scale_factor = 1 / (1 - mwh_invalid_bins / pre_mwh)
    else:
        missing_bins_unc_scale_factor = 1

    pp_daterange = (
        cfg.analysis_last_dt_utc_start + pd.Timedelta(seconds=cfg.timebase_s)
    ) - cfg.analysis_first_dt_utc_start
    pp_possible_hours = pp_daterange.total_seconds() / 3600

    pp_df["pw_at_mid_expected"] = pp_df["pw_at_mid_post"]
    pp_df["pw_sem_at_mid_expected"] = pp_df["pw_sem_at_mid_post"]
    pp_df.loc[~pp_df["is_invalid_bin"], "pw_at_mid_expected"] = pp_df.loc[~pp_df["is_invalid_bin"], "pw_at_mid_pre"]
    pp_df.loc[~pp_df["is_invalid_bin"], "pw_sem_at_mid_expected"] = pp_df.loc[
        ~pp_df["is_invalid_bin"],
        "pw_sem_at_mid_pre",
    ]

    expected_post_mwh = (pp_df["pw_at_mid_expected"] * pp_df["hours_for_mwh_calc"]).sum() / 1000
    expected_post_se_mwh = (
        pp_df["hours_for_mwh_calc"].sum() * np.sqrt(((pp_df["f"] * pp_df["pw_sem_at_mid_expected"]) ** 2).sum()) / 1000
    )

    post_mwh = (pp_df["pw_at_mid_post"] * pp_df["hours_for_mwh_calc"]).sum() / 1000
    post_se_mwh = (
        pp_df["hours_for_mwh_calc"].sum() * np.sqrt(((pp_df["f"] * pp_df["pw_sem_at_mid_post"]) ** 2).sum()) / 1000
    )

    uplift_mwh = post_mwh - expected_post_mwh

    uplift_se_mwh = np.sqrt(expected_post_se_mwh**2 + post_se_mwh**2)

    valid_count = round(
        np.minimum(pp_df["hours_pre"].sum(), pp_df["hours_post"].sum()) * 3600 / cfg.timebase_s,
    )
    p_low = (1 - confidence_level) / 2
    p_high = 1 - p_low
    t_value_one_sigma = t.ppf(
        norm.cdf(1),
        valid_count - 1,
    )
    t_values = t.ppf(
        p_high,
        np.minimum(pp_df["hours_pre"].clip(lower=2), pp_df["hours_post"].clip(lower=2)) * 3600 / cfg.timebase_s - 1,
    )
    unc_one_sigma_mwh = uplift_se_mwh * t_value_one_sigma

    pp_df = add_uplift_cols_to_pp_df(pp_df, p_low=p_low, p_high=p_high, t_values=t_values)

    pp_valid_hours_pre = pp_df["hours_pre"].sum()
    pp_valid_hours_post = pp_df["hours_post"].sum()
    pp_valid_hours = pp_valid_hours_pre + pp_valid_hours_post

    pp_results = {
        "time_calculated": pd.Timestamp.now("UTC"),
        "uplift_frc": uplift_mwh / expected_post_mwh,
        "unc_one_sigma_frc": unc_one_sigma_mwh / expected_post_mwh * missing_bins_unc_scale_factor,
        "t_value_one_sigma": t_value_one_sigma,
        "missing_bins_unc_scale_factor": missing_bins_unc_scale_factor,
        "pp_valid_hours_pre": pp_valid_hours_pre,
        "pp_valid_hours_post": pp_valid_hours_post,
        "pp_valid_hours": pp_valid_hours,
        "pp_data_coverage": pp_valid_hours / pp_possible_hours,
        "pp_invalid_bin_count": pp_df["is_invalid_bin"].sum(),
    }

    if plot_cfg is not None:
        plot_pre_post_pp_analysis(
            test_name=test_name,
            ref_name=ref_name,
            pp_df=pp_df,
            pre_df=pre_df,
            post_df=post_df,
            ws_col=ws_col,
            pw_col=pw_col,
            wd_col=wd_col,
            timebase_s=cfg.timebase_s,
            plot_cfg=plot_cfg,
            confidence_level=confidence_level,
        )

        if test_df is not None:
            test_df_pp_period = test_df[
                cfg.analysis_first_dt_utc_start : cfg.analysis_last_dt_utc_start  # type: ignore[misc]
            ]
            plot_pp_data_coverage(
                test_name=test_name,
                ref_name=ref_name,
                pp_df=pp_df,
                test_df_pp_period=test_df_pp_period,
                ws_bin_width=cfg.ws_bin_width,
                timebase_s=cfg.timebase_s,
                plot_cfg=plot_cfg,
            )

    return pp_results, pp_df


def calc_power_only_and_reversed_uplifts(
    *,
    cfg: WindUpConfig,
    test_wtg: Turbine,
    ref_name: str,
    lt_df: pd.DataFrame,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    pw_col: str,
    wd_col: str,
    confidence_level: float = 0.9,
) -> tuple[float, float]:
    # calculate power only forward result
    pre_power_only = pre_df.copy()
    pre_power_only["ref_ws_power_only_detrended"] = (
        pre_power_only["ref_ws_est_from_power_only"] * pre_power_only["ws_rom"]
    )
    post_power_only = post_df.copy()
    post_power_only["ref_ws_power_only_detrended"] = (
        post_power_only["ref_ws_est_from_power_only"] * post_power_only["ws_rom"]
    )
    power_only_results, _ = pre_post_pp_analysis(
        cfg=cfg,
        test_wtg=test_wtg,
        ref_name=ref_name,
        lt_df=lt_df,
        pre_df=pre_power_only,
        post_df=post_power_only,
        ws_col="ref_ws_power_only_detrended",
        pw_col=pw_col,
        wd_col=wd_col,
        plot_cfg=None,
        confidence_level=confidence_level,
    )

    # need to predict the reference wind speed using the test wind speed for reverse analysis
    pre_power_only["test_ws_power_only_detrended"] = (
        pre_power_only["test_ws_est_from_power_only"] / pre_power_only["ws_rom"]
    )
    post_power_only["test_ws_power_only_detrended"] = (
        post_power_only["test_ws_est_from_power_only"] / post_power_only["ws_rom"]
    )
    reversed_results, _ = pre_post_pp_analysis(
        cfg=cfg,
        test_wtg=test_wtg,
        ref_name=ref_name,
        lt_df=lt_df,
        pre_df=pre_power_only,
        post_df=post_power_only,
        ws_col="ref_ws_power_only_detrended",
        pw_col=pw_col,
        wd_col=wd_col,
        plot_cfg=None,
        confidence_level=confidence_level,
        reverse=True,
    )

    poweronly_uplift_frc = power_only_results["uplift_frc"]
    reversed_uplift_frc = reversed_results["uplift_frc"]
    return poweronly_uplift_frc, reversed_uplift_frc


def pre_post_pp_analysis_with_reversal(
    *,
    cfg: WindUpConfig,
    test_wtg: Turbine,
    ref_name: str,
    lt_df: pd.DataFrame | None,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    ws_col: str,
    pw_col: str,
    wd_col: str,
    plot_cfg: PlotConfig | None,
    confidence_level: float = 0.9,
    test_df: pd.DataFrame | None = None,
) -> tuple[dict, pd.DataFrame]:
    pp_results, pp_df = pre_post_pp_analysis(
        cfg=cfg,
        test_wtg=test_wtg,
        ref_name=ref_name,
        lt_df=lt_df,
        pre_df=pre_df,
        post_df=post_df,
        ws_col=ws_col,
        pw_col=pw_col,
        wd_col=wd_col,
        plot_cfg=plot_cfg,
        confidence_level=confidence_level,
        test_df=test_df,
    )

    if test_wtg.name == ref_name:
        poweronly_uplift_frc = np.nan
        reversed_uplift_frc = np.nan
        reversal_error = 0.0
    else:
        poweronly_uplift_frc, reversed_uplift_frc = calc_power_only_and_reversed_uplifts(
            cfg=cfg,
            test_wtg=test_wtg,
            ref_name=ref_name,
            lt_df=lt_df,
            pre_df=pre_df,
            post_df=post_df,
            pw_col=pw_col,
            wd_col=wd_col,
            confidence_level=confidence_level,
        )
        reversal_error = reversed_uplift_frc - poweronly_uplift_frc
    if plot_cfg is not None:
        logger.info(f"\nresults for test={test_wtg.name} ref={ref_name}:\n")
        logger.info(f"hours pre = {pp_results['pp_valid_hours_pre']:.1f}")
        logger.info(f"hours post = {pp_results['pp_valid_hours_post']:.1f}")
        logger.info(f"\nuplift estimate before adjustments = {100*pp_results['uplift_frc']:.1f} %")

        logger.info(f"\npower only uplift estimate = {100 * poweronly_uplift_frc:.1f} %")
        logger.info(f"reversed (power only) uplift estimate = {100 * reversed_uplift_frc:.1f} %\n")

    pp_results["uplift_noadj_frc"] = pp_results["uplift_frc"]
    pp_results["unc_one_sigma_noadj_frc"] = pp_results["unc_one_sigma_frc"]

    pp_results["poweronly_uplift_frc"] = poweronly_uplift_frc
    pp_results["reversed_uplift_frc"] = reversed_uplift_frc
    pp_results["reversal_error"] = reversal_error
    pp_results["uplift_frc"] = pp_results["uplift_noadj_frc"] + reversal_error / 2
    pp_results["unc_one_sigma_lowerbound_frc"] = abs(reversal_error) / 2
    pp_results["unc_one_sigma_frc"] = max(
        pp_results["unc_one_sigma_noadj_frc"],
        pp_results["unc_one_sigma_lowerbound_frc"],
    )

    return pp_results, pp_df


def pre_post_pp_analysis_with_reversal_and_bootstrapping(
    *,
    cfg: WindUpConfig,
    test_wtg: Turbine,
    ref_name: str,
    lt_df: pd.DataFrame | None,
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    ws_col: str,
    pw_col: str,
    wd_col: str,
    plot_cfg: PlotConfig | None,
    random_seed: int,
    confidence_level: float = 0.9,
    test_df: pd.DataFrame | None = None,
) -> tuple[dict, pd.DataFrame]:
    pp_results, pp_df = pre_post_pp_analysis_with_reversal(
        cfg=cfg,
        test_wtg=test_wtg,
        ref_name=ref_name,
        lt_df=lt_df,
        pre_df=pre_df,
        post_df=post_df,
        ws_col=ws_col,
        pw_col=pw_col,
        wd_col=wd_col,
        plot_cfg=plot_cfg,
        test_df=test_df,
    )

    pre_df_dropna = pre_df.dropna(subset=[ws_col, pw_col, wd_col])
    post_df_dropna = post_df.dropna(subset=[ws_col, pw_col, wd_col])

    n_samples = cfg.bootstrap_runs_override if cfg.bootstrap_runs_override else round(40 * (1 / (1 - confidence_level)))
    if plot_cfg is not None:
        logger.info(f"Running block bootstrapping uncertainty analysis n_samples = {n_samples}")
    bootstrapped_uplifts = np.empty(n_samples)
    bootstrapped_uplifts[:] = np.nan
    rng = np.random.default_rng(seed=random_seed)
    for n in tqdm(range(n_samples)):
        num_blocks = rng.choice([9, 10, 11])
        block_size_pre = math.floor(len(pre_df_dropna) / num_blocks)
        block_size_post = math.floor(len(post_df_dropna) / num_blocks)
        # randomly remove rows to make the dataframes the perfect length
        pre_target_len = num_blocks * block_size_pre
        post_target_len = num_blocks * block_size_post
        pre_rows_to_keep = rng.choice(pre_df_dropna.index, size=pre_target_len, replace=False)
        post_rows_to_keep = rng.choice(post_df_dropna.index, size=post_target_len, replace=False)
        pre_df_trim = pre_df_dropna.loc[pre_rows_to_keep].sort_index()
        post_df_trim = post_df_dropna.loc[post_rows_to_keep].sort_index()

        blocks = rng.choice(num_blocks, size=num_blocks, replace=True)

        pre_sample_ilocs = (block_size_pre * blocks[:, None] + np.arange(block_size_pre)[None, :]).flatten()
        pre_df_ = pre_df_trim.iloc[pre_sample_ilocs].sort_index()

        post_sample_ilocs = (block_size_post * blocks[:, None] + np.arange(block_size_post)[None, :]).flatten()
        post_df_ = post_df_trim.iloc[post_sample_ilocs].sort_index()

        try:
            sample_results, _ = pre_post_pp_analysis_with_reversal(
                cfg=cfg,
                test_wtg=test_wtg,
                ref_name=ref_name,
                lt_df=lt_df,
                pre_df=pre_df_.reset_index(),
                post_df=post_df_.reset_index(),
                ws_col=ws_col,
                pw_col=pw_col,
                wd_col=wd_col,
                plot_cfg=None,
            )
            bootstrapped_uplifts[n] = sample_results["uplift_frc"]
        except RuntimeError:
            result_manager.warning(f"RuntimeError on sample {n}")
            bootstrapped_uplifts[n] = np.nan

    if np.isnan(bootstrapped_uplifts).sum() < 0.5 * len(bootstrapped_uplifts):
        median = float(np.nanmedian(bootstrapped_uplifts))
        lower = float(np.nanpercentile(bootstrapped_uplifts, 100 * (1 - confidence_level) / 2))
        upper = float(np.nanpercentile(bootstrapped_uplifts, 100 * (1 - (1 - confidence_level) / 2)))
        unc_one_sigma = (upper - lower) / 2 / norm.ppf((1 + confidence_level) / 2)
    else:
        median = np.nan
        lower = np.nan
        upper = np.nan
        unc_one_sigma = np.nan

    if plot_cfg is not None:
        msg = (
            f"block bootstrapping uncertainty analysis results (conf={100*confidence_level:.0f}%):"
            f"\n  median = {100*median:.1f} %"
            f"\n  lower = {100*lower:.1f} %"
            f"\n  upper = {100*upper:.1f} %"
            f"\n  unc_one_sigma = {100*unc_one_sigma:.1f} %"
        )
        logger.info(msg)

    pp_results["unc_one_sigma_bootstrap_frc"] = unc_one_sigma
    pp_results["unc_one_sigma_frc"] = max(
        pp_results["unc_one_sigma_frc"],
        pp_results["unc_one_sigma_bootstrap_frc"],
    )

    p_low = (1 - confidence_level) / 2
    p_high = 1 - p_low

    pp_results[f"uplift_p{p_low * 100:.0f}_frc"] = pp_results["uplift_frc"] + pp_results[
        "unc_one_sigma_frc"
    ] * norm.ppf((1 + confidence_level) / 2)
    pp_results[f"uplift_p{p_high * 100:.0f}_frc"] = pp_results["uplift_frc"] - pp_results[
        "unc_one_sigma_frc"
    ] * norm.ppf((1 + confidence_level) / 2)
    if plot_cfg is not None:
        logger.info(f"\ncat A 1 sigma unc = {100 * pp_results['unc_one_sigma_noadj_frc']:.1f} %")
        if pp_results["unc_one_sigma_lowerbound_frc"] > 0.05 / 100:
            logger.info(f"abs reversal error / 2 = {100 * pp_results['unc_one_sigma_lowerbound_frc']:.1f} %")
        else:
            logger.info(f"abs reversal error / 2 = {100 * pp_results['unc_one_sigma_lowerbound_frc']:.3f} %")
        logger.info(f"bootstrap 1 sigma unc = {100 * pp_results['unc_one_sigma_bootstrap_frc']:.1f} %")
        logger.info(f"missing bins scale factor = {pp_results['missing_bins_unc_scale_factor']:.3f}")
        logger.info(f"final 1 sigma unc = {100 * pp_results['unc_one_sigma_frc']:.1f} %\n")

        logger.info(f"final uplift estimate = {100*pp_results['uplift_frc']:.1f} %")
        logger.info(f"final P95 uplift estimate = {100*pp_results[f'uplift_p{p_high * 100:.0f}_frc']:.1f} %")
        logger.info(f"final P5 uplift estimate = {100*pp_results[f'uplift_p{p_low * 100:.0f}_frc']:.1f} %")

    return pp_results, pp_df



from pathlib import Path

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from wind_up.detrend import apply_wsratio_v_wd_scen, calc_wsratio_v_wd_scen, check_applied_detrend
from wind_up.models import WindUpConfig
from wind_up.smart_data import add_smart_lat_long_to_cfg


def test_apply_wsratio_v_wd_scen_pre() -> None:
    expected_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_LSA_T12_pre_df.parquet")
    pre_df = expected_df.drop(columns=["ws_rom", "ref_ws_detrended"])
    wsratio_v_dir_scen = pd.read_parquet(
        Path(__file__).parents[0] / "test_data/LSA_T13_LSA_T12_wsratio_v_dir_scen.parquet",
    )
    ref_ws_col = "ref_ws_est_blend"
    ref_wd_col = "ref_YawAngleMean"
    actual_df = apply_wsratio_v_wd_scen(pre_df, wsratio_v_dir_scen, ref_ws_col=ref_ws_col, ref_wd_col=ref_wd_col)
    assert_frame_equal(actual_df, expected_df)


def test_apply_wsratio_v_wd_scen_post() -> None:
    expected_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_LSA_T12_post_df.parquet")
    post_df = expected_df.drop(columns=["ws_rom", "ref_ws_detrended"])
    wsratio_v_dir_scen = pd.read_parquet(
        Path(__file__).parents[0] / "test_data/LSA_T13_LSA_T12_wsratio_v_dir_scen.parquet",
    )
    ref_ws_col = "ref_ws_est_blend"
    ref_wd_col = "ref_YawAngleMean"
    actual_df = apply_wsratio_v_wd_scen(post_df, wsratio_v_dir_scen, ref_ws_col=ref_ws_col, ref_wd_col=ref_wd_col)
    assert_frame_equal(actual_df, expected_df)


def test_check_applied_detrend(test_lsa_t13_config: WindUpConfig) -> None:
    cfg = test_lsa_t13_config
    pre_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_LSA_T12_pre_df.parquet")
    post_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_LSA_T12_post_df.parquet")
    test_name = "LSA_T13"
    ref_name = "LSA_T12"
    ref_wtg = next(x for x in cfg.asset.wtgs if x.name == ref_name)
    ref_lat = ref_wtg.latitude
    ref_long = ref_wtg.longitude
    test_ws_col = "test_ws_est_blend"
    ref_ws_col = "ref_ws_est_blend"
    ref_wd_col = "ref_YawAngleMean"
    detrend_ws_col = "ref_ws_detrended"

    detrend_pre_r2_improvement, detrend_post_r2_improvement = check_applied_detrend(
        test_name=test_name,
        ref_name=ref_name,
        ref_lat=ref_lat,
        ref_long=ref_long,
        pre_df=pre_df,
        post_df=post_df,
        test_ws_col=test_ws_col,
        ref_ws_col=ref_ws_col,
        detrend_ws_col=detrend_ws_col,
        ref_wd_col=ref_wd_col,
        cfg=cfg,
        plot_cfg=None,
    )

    assert detrend_pre_r2_improvement == pytest.approx(0.03464757681986863)
    assert detrend_post_r2_improvement == pytest.approx(0.03776561982402227)


def test_calc_wsratio_v_wd_scen(test_lsa_t13_config: WindUpConfig) -> None:
    # this test case borrows logic and results from check_applied_detrend where data which has already been detrended
    # is used to calculate the wsratio_v_wd_scen again to check it is flat
    cfg = test_lsa_t13_config
    md_df = pd.read_csv(Path(__file__).parents[0] / "test_data/smart_data/Lisa Wind Farm/Lisa Wind Farm_md.csv")
    cfg = add_smart_lat_long_to_cfg(md=md_df, cfg=cfg)
    pre_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_LSA_T12_pre_df.parquet")
    test_name = "LSA_T13"
    ref_name = "LSA_T12"
    ref_wtg = next(x for x in cfg.asset.wtgs if x.name == ref_name)
    ref_lat = ref_wtg.latitude
    ref_long = ref_wtg.longitude
    test_ws_col = "test_ws_est_blend"
    ref_wd_col = "ref_YawAngleMean"
    detrend_ws_col = "ref_ws_detrended"

    expected_pre_df = pd.read_parquet(
        Path(__file__).parents[0] / "test_data/LSA_T13_LSA_T12_check_pre_wsratio_v_dir_scen.parquet",
    )
    actual_pre_df = calc_wsratio_v_wd_scen(
        test_name=test_name,
        ref_name=ref_name,
        ref_lat=ref_lat,
        ref_long=ref_long,
        detrend_df=pre_df,
        test_ws_col=test_ws_col,
        ref_ws_col=detrend_ws_col,  # different intentionally
        ref_wd_col=ref_wd_col,
        cfg=cfg,
        plot_cfg=None,
    )

    assert_frame_equal(actual_pre_df, expected_pre_df)



from pathlib import Path

import pandas as pd
from pandas.testing import assert_frame_equal

from wind_up.models import WindUpConfig
from wind_up.scada_power_curve import calc_pc_and_rated_ws
from wind_up.ws_est import add_ws_est


def test_ws_est(test_homer_config: WindUpConfig) -> None:
    cfg = test_homer_config
    wf_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/Homer Wind Farm_July2023_scada_improved.parquet")

    # usually x_bin_width would be cfg.ws_bin_width / 2 but that gives invalid power curve with so little data
    pc_per_ttype, rated_ws_per_ttype = calc_pc_and_rated_ws(
        cfg=cfg,
        wf_df=wf_df,
        x_col="WindSpeedMean",
        y_col="pw_clipped",
        x_bin_width=cfg.ws_bin_width * 2.01,
        plot_cfg=None,
    )
    adf = add_ws_est(cfg=cfg, wf_df=wf_df, pc_per_ttype=pc_per_ttype, plot_cfg=None)
    edf = pd.read_parquet(Path(__file__).parents[0] / "test_data/test_add_ws_est.parquet")
    assert_frame_equal(adf, edf)



import matplotlib.pyplot as plt
import pandas as pd

from wind_up.constants import (
    REANALYSIS_WD_COL,
    REANALYSIS_WS_COL,
    SCATTER_ALPHA,
    SCATTER_S,
    TIMESTAMP_COL,
)
from wind_up.models import PlotConfig


def plot_find_best_shift_and_corr(
    wf_ws_df: pd.DataFrame,
    *,
    reanalysis_df: pd.DataFrame,
    shifts: list[int],
    corrs: list[float],
    wf_name: str,
    datastream_id: str,
    best_corr: float,
    best_s: int,
    timebase_s: int,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure()
    plt.plot(shifts, corrs, marker=".")
    plot_title = f"correlation vs shift for {datastream_id}"
    plt.title(plot_title)
    plt.xlabel(f"shift [{timebase_s // 60:.0f}min rows]")
    plt.ylabel("correlation with wind farm mean wind speed")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")
    plt.close()

    plot_df = wf_ws_df.merge(reanalysis_df[REANALYSIS_WS_COL].shift(best_s), left_index=True, right_index=True)
    plt.figure()
    plt.scatter(plot_df[REANALYSIS_WS_COL], plot_df["WindSpeedMean"], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plot_title = f"{wf_name} vs {datastream_id}"
    plt.suptitle(plot_title, fontsize=14)
    plt.title(f"corr= {best_corr:.3f}, shift = {best_s}", fontsize=10)
    plt.xlabel("reanalysis wind speed [m/s]")
    plt.ylabel("wind farm mean wind speed [m/s]")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")
    plt.close()


def plot_wf_and_reanalysis_sample_timeseries(wf_df: pd.DataFrame, plot_cfg: PlotConfig) -> None:
    max_timestamp = (
        wf_df.dropna(subset=["WindSpeedMean", "YawAngleMean", REANALYSIS_WS_COL, REANALYSIS_WD_COL])
        .index.get_level_values(TIMESTAMP_COL)
        .max()
    )
    one_month_before_max = max_timestamp - pd.Timedelta(days=365.25 / 12)
    filt_df = wf_df[wf_df.index.get_level_values(TIMESTAMP_COL) >= one_month_before_max]
    plt.figure(figsize=(10, 6))
    wtg_names = filt_df.index.unique(level="TurbineName")
    for wtg_name in wtg_names:
        plt.plot(filt_df.loc[wtg_name, "WindSpeedMean"], label=wtg_name)
    plt.plot(filt_df.loc[wtg_name, REANALYSIS_WS_COL], label="reanalysis", linewidth=3)
    plt.legend(loc="upper left", ncol=max(1, round(len(wtg_names) // 5)), fontsize="small")
    plot_title = "turbine and reanalysis wind speeds"
    plt.title(plot_title)
    plt.xticks(rotation=90)
    plt.xlabel("datetime")
    plt.ylabel("wind speed [m/s]")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")
    plt.close()

    plt.figure(figsize=(10, 6))
    wtg_names = filt_df.index.unique(level="TurbineName")
    for wtg_name in wtg_names:
        plt.plot(filt_df.loc[wtg_name, "YawAngleMean"], label=wtg_name)
    plt.plot(filt_df.loc[wtg_name, REANALYSIS_WD_COL], label="reanalysis", linewidth=3)
    plt.legend(loc="lower left", ncol=max(1, round(len(wtg_names) // 5)), fontsize="small")
    plot_title = "turbine and reanalysis wind directions"
    plt.title(plot_title)
    plt.xticks(rotation=90)
    plt.xlabel("datetime")
    plt.ylabel("wind direction [deg]")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")
    plt.close()



from pathlib import Path

import pandas as pd
import pytest

from wind_up.constants import REANALYSIS_WS_COL
from wind_up.models import WindUpConfig
from wind_up.windspeed_drift import check_windspeed_drift


def test_check_windspeed_drift(test_lsa_t13_config: WindUpConfig) -> None:
    cfg = test_lsa_t13_config
    test_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_test_df.parquet")
    test_name = "LSA_T13"
    test_ws_col = "test_ws_est_blend"
    test_max_ws_drift, test_max_ws_drift_pp_period = check_windspeed_drift(
        wtg_df=test_df,
        wtg_name=test_name,
        ws_col=test_ws_col,
        reanalysis_ws_col="test_" + REANALYSIS_WS_COL,
        cfg=cfg,
        plot_cfg=None,
    )
    assert test_max_ws_drift == pytest.approx(0.45289044075068974)
    assert test_max_ws_drift_pp_period == pytest.approx(0.42913942378401204)



import datetime as dt

import pandas as pd
import pytest

from tests.conftest import TEST_DATA_FLD
from wind_up.constants import TIMESTAMP_COL
from wind_up.models import WindUpConfig
from wind_up.smart_data import (
    add_smart_lat_long_to_cfg,
    calc_last_xmin_datetime_in_month,
    calc_month_list_and_time_info,
    load_smart_md_from_file,
    load_smart_scada_and_md_from_file,
)

TIMEBASE_PD_TIMEDELTA = pd.Timedelta("10min")


def test_calc_last_xmin_datetime_in_month() -> None:
    inputs = [
        dt.datetime(2020, 1, 13),
        dt.datetime(2020, 2, 1),
        dt.datetime(2020, 2, 1, 0, 0),
        dt.datetime(2020, 2, 1, 0, 10),
        dt.datetime(2020, 2, 29, 23, 50),
    ]
    expected = [
        dt.datetime(2020, 1, 31, 23, 50),
        dt.datetime(2020, 2, 29, 23, 50),
        dt.datetime(2020, 2, 29, 23, 50),
        dt.datetime(2020, 2, 29, 23, 50),
        dt.datetime(2020, 2, 29, 23, 50),
    ]
    for i, e in zip(inputs, expected, strict=True):
        assert calc_last_xmin_datetime_in_month(i, TIMEBASE_PD_TIMEDELTA) == pd.Timestamp(e)


def test_add_smart_lat_long_to_cfg(test_marge_config: WindUpConfig) -> None:
    cfg = test_marge_config
    md_df = load_smart_md_from_file(asset_name="Marge Wind Farm", test_mode=True)
    cfg = add_smart_lat_long_to_cfg(md=md_df, cfg=cfg)
    assert cfg.asset.wtgs[0].latitude == pytest.approx(-53.741243623)
    assert cfg.asset.wtgs[0].longitude == pytest.approx(10.0580049240437)
    assert cfg.asset.wtgs[-1].latitude == pytest.approx(-53.7496477908)
    assert cfg.asset.wtgs[-1].longitude == pytest.approx(10.0752109226079)


def test_calc_month_list_and_time_info() -> None:
    md_df = load_smart_md_from_file(asset_name="Marge Wind Farm", test_mode=True)
    month_start_list_no_tz, last_smart_dt_no_tz, smart_tz, smart_tf = calc_month_list_and_time_info(
        asset_name="Marge Wind Farm",
        first_datetime_utc_start=pd.Timestamp("2019-12-25 00:20:00", tz="UTC"),
        last_datetime_utc_start=pd.Timestamp("2020-02-28 23:50:00", tz="UTC"),
        md=md_df,
        timebase_s=600,
    )
    assert month_start_list_no_tz == [
        pd.Timestamp("2019-12-25 00:00:00"),
        pd.Timestamp("2020-01-01 00:00:00"),
        pd.Timestamp("2020-02-01 00:00:00"),
    ]
    assert last_smart_dt_no_tz == pd.Timestamp("2020-02-29 00:00:00")
    assert smart_tz == "UTC"
    assert smart_tf == "End"


def test_load_smart_scada_and_md_from_file() -> None:
    test_data_dir = TEST_DATA_FLD / "smart_data" / "Marge Wind Farm"
    first_datetime_utc_start = pd.Timestamp("2020-02-26 23:50:00", tz="UTC")
    last_datetime_utc_start = pd.Timestamp("2020-02-29 23:40:00", tz="UTC")
    scada_raw, md = load_smart_scada_and_md_from_file(
        asset_name="Marge Wind Farm",
        scada_df=pd.concat([pd.read_parquet(i) for i in test_data_dir.glob("*.parquet")]),
        metadata_df=pd.read_csv(test_data_dir / "Marge Wind Farm_md.csv"),
        first_datetime_utc_start=first_datetime_utc_start,
        last_datetime_utc_start=last_datetime_utc_start,
        timebase_s=600,
    )
    assert len(scada_raw) == 3 * 9 * 24 * 6  # 3 days, 9 turbines
    assert scada_raw.index.name == TIMESTAMP_COL
    assert scada_raw.index.min() == first_datetime_utc_start
    assert scada_raw.index.max() == last_datetime_utc_start
    assert scada_raw["smart_dtTimeStamp"].min() == pd.Timestamp("2020-02-27 00:00:00")
    assert scada_raw["smart_dtTimeStamp"].max() == pd.Timestamp("2020-02-29 23:50:00")
